/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */


#ifndef __nanojit_NativeARM__
#define __nanojit_NativeARM__

#ifdef PERFM
#include "../vprof/vprof.h"
#define count_instr() _nvprof("arm",1)
#define count_prolog() _nvprof("arm-prolog",1); count_instr();
#define count_imt() _nvprof("arm-imt",1) count_instr()
#else
#define count_instr()
#define count_prolog()
#define count_imt()
#endif

#if defined(NANOJIT_DEBUG)
#define ARM_ARCH_AT_LEAST(wanted) (_config.arm_arch >= (wanted))
#define ARM_VFP (_config.arm_vfp)
#else
/* Note: Non DEBUG builds will ignore arm_arch if it is lower than
   NANOJIT_COMPILER_ARM_ARCH, and will ignore arm_vfp if NANOJIT_COMPILER_ARM_ARCH
   is greater or equal to 7. */
#define ARM_ARCH_AT_LEAST(wanted) \
	((NANOJIT_COMPILER_ARM_ARCH >= (wanted)) || (_config.arm_arch >= (wanted)))
#define ARM_VFP ((NANOJIT_COMPILER_ARM_ARCH >= 7) || (_config.arm_vfp))
#endif

namespace nanojit
{


// GCC defines __ARM_PCS_VFP if it uses hardware floating point ABI
// See http://gcc.gnu.org/viewcvs?view=revision&revision=162637
#ifdef __ARM_PCS_VFP
#  define NANOJIT_ARM_EABI_HARD_FLOAT 1
#endif

#ifdef NANOJIT_ARM_EABI_HARD_FLOAT
#  define ARM_EABI_HARD true
#else
#  define ARM_EABI_HARD false
#endif

// only d0-d7 are used; in addition, we'll use d0 as s0-s1 for i2d/u2f/etc.
#define NANOJIT_VFP_MAX_REGISTERS            8
#define NANOJIT_MAX_REGISTERS                (11 + NANOJIT_VFP_MAX_REGISTERS)
#define NANOJIT_MAX_STACK_ENTRY              1024
#define NANOJIT_MAX_PARAMETERS               16
#define NANOJIT_ALIGN_STACK                  8

#define NANOJIT_JTBL_SUPPORTED               1
#define NANOJIT_EXPANDED_LOADSTORE_SUPPORTED 1
#define NANOJIT_F2I_SUPPORTED                1
#define NANOJIT_SOFTFLOAT_SUPPORTED          1
#define NANOJIT_DIVI_SUPPORTED               0

#define NANOJIT_JTBL_ALLOWED_IDX_REGS        GpRegs

//#define RA_REGISTERS_OVERLAP            0

#if defined(RA_REGISTERS_OVERLAP)
#define firstAvailableReg               getAvailableReg
#define getFatherReg                    _allocator.getSuitableRegFor
#else
#define NANOJIT_RA_PREFERS_MSREG
#endif
}

#include "NativeCommon.h"

namespace nanojit
{

typedef int NIns;

// Bytes of icache to flush after Assembler::patch
const size_t LARGEST_BRANCH_PATCH = 2 * sizeof(NIns);

/* ARM registers */
static const Register
	R0  = { 0 },
	R1  = { 1 },
	R2  = { 2 },
	R3  = { 3 },
	R4  = { 4 },
	R5  = { 5 },
	R6  = { 6 },
	R7  = { 7 },
	R8  = { 8 },
	R9  = { 9 },
	R10 = { 10 },
	R11 = { 11 },
	R12 = { 12 },
	R13 = { 13 },
	R14 = { 14 },
	R15 = { 15 },

	FP  = { 11 },
	IP  = { 12 }, // will be trashed by calls
	SP  = { 13 },
	LR  = { 14 },
	PC  = { 15 },

	// VFP regs
	S0  = { 16 },
	S1  = { REGNUM(S0) + 1 },
	S2  = { REGNUM(S0) + 2 },
	S3  = { REGNUM(S0) + 3 },
	S4  = { REGNUM(S0) + 4 },
	S5  = { REGNUM(S0) + 5 },
	S6  = { REGNUM(S0) + 6 },
	S7  = { REGNUM(S0) + 7 },
	S8  = { REGNUM(S0) + 8 },
	S9  = { REGNUM(S0) + 9 },
	S10 = { REGNUM(S0) +10 },
	S11 = { REGNUM(S0) +11 },
	S12 = { REGNUM(S0) +12 },
	S13 = { REGNUM(S0) +13 },
	S14 = { REGNUM(S0) +14 },
	S15 = { REGNUM(S0) +15 },
	S16 = { REGNUM(S0) +16 },
	S17 = { REGNUM(S0) +17 },
	S18 = { REGNUM(S0) +18 },
	S19 = { REGNUM(S0) +19 },
	S20 = { REGNUM(S0) +20 },
	S21 = { REGNUM(S0) +21 },
	S22 = { REGNUM(S0) +22 },
	S23 = { REGNUM(S0) +23 },
	S24 = { REGNUM(S0) +24 },
	S25 = { REGNUM(S0) +25 },
	S26 = { REGNUM(S0) +26 },
	S27 = { REGNUM(S0) +27 },
	S28 = { REGNUM(S0) +28 },
	S29 = { REGNUM(S0) +29 },
	S30 = { REGNUM(S0) +30 },
	S31 = { REGNUM(S0) +31 },
	// We reverse the order of the Dn register banks: first, the Dn regs that have no Sn subregs
	D16 = { 48 },
	D17 = { REGNUM(D16) + 1 },
	D18 = { REGNUM(D16) + 2 },
	D19 = { REGNUM(D16) + 3 },
	D20 = { REGNUM(D16) + 4 },
	D21 = { REGNUM(D16) + 5 },
	D22 = { REGNUM(D16) + 6 },
	D23 = { REGNUM(D16) + 7 },
	D24 = { REGNUM(D16) + 8 },
	D25 = { REGNUM(D16) + 9 },
	D26 = { REGNUM(D16) +10 },
	D27 = { REGNUM(D16) +11 },
	D28 = { REGNUM(D16) +12 },
	D29 = { REGNUM(D16) +13 },
	D30 = { REGNUM(D16) +14 },
	D31 = { REGNUM(D16) +15 },
	// The following regs are just alternate names for gruops of registers already listed above
	D0  = { 64 },
	D1  = { REGNUM(D0) + 1 },
	D2  = { REGNUM(D0) + 2 },
	D3  = { REGNUM(D0) + 3 },
	D4  = { REGNUM(D0) + 4 },
	D5  = { REGNUM(D0) + 5 },
	D6  = { REGNUM(D0) + 6 },
	D7  = { REGNUM(D0) + 7 },
	D8  = { REGNUM(D0) + 8 },
	D9  = { REGNUM(D0) + 9 },
	D10 = { REGNUM(D0) +10 },
	D11 = { REGNUM(D0) +11 },
	D12 = { REGNUM(D0) +12 },
	D13 = { REGNUM(D0) +13 },
	D14 = { REGNUM(D0) +14 },
	D15 = { REGNUM(D0) +15 },
	
	Q0  = { 80 },
	Q1  = { 81 },
	Q2  = { 82 },
	Q3  = { 83 },
	Q4  = { 84 },
	Q5  = { 85 },
	Q6  = { 86 },
	Q7  = { 87 },
	Q8  = { 88 },
	Q9  = { 89 },
	Q10 = { 90 },
	Q11 = { 91 },
	Q12 = { 92 },
	Q13 = { 93 },
	Q14 = { 94 },
	Q15 = { 95 },

	FirstDFloatReg = D16,
	FirstQFloatReg = Q0,
	FirstSFloatReg = S0,

	UnspecifiedReg = { 128 },
	deprecated_UnknownReg = { 128 },     // XXX: remove eventually, see bug 538924

	SBZ = { 0 },       // Used for 'should-be-zero' fields in instructions with
					   // unused register fields.
	SBO = { 0xF } ;    // Used for 'should-be-one' fields in instructions with
                       // unused register fields.


static const uint32_t FirstRegNum = 0; /* R0 */
static const uint32_t LastRegNum = 95; /* Q15 */ 

typedef uint64_t RegisterMask;

#if defined(RA_REGISTERS_OVERLAP)
#define rmask(r) ARM_REG_MASKS[REGNUM(r)]
extern const RegisterMask ARM_REG_MASKS[LastRegNum+1];
#endif


// Some VFP CPUs (non-NEON) lack the full complement of 32 FP registers, implementing only S0-S15.
// Since we do not dynamically configure for the ARM variant in use, we must compile for the
// lowest common denominator among supported targets, currently VFP-D16 (Tegra-2).
// FIXME: We should identify the architectural variant at runtime. See bugs 718811 and 704111.

const RegisterMask FpSRegs = 0x00000000ffff0000LL; // S0-S15
const RegisterMask FpDRegs = 0x00000000ffff0000LL; // D0-D7
const RegisterMask FpQRegs = 0x00000000ffff0000LL; // Q0-Q3

const RegisterMask GpRegs  = 0x000000000000ffffLL; // R0-R10,FP,IP,SP,LR,PC
const RegisterMask FpRegs  = 0x00000000ffff0000LL; // all floating point registers


/* ARM condition codes */
typedef enum {
	EQ = 0x0, // Equal
	NE = 0x1, // Not Equal
	CS = 0x2, // Carry Set (or HS)
	HS = 0x2,
	CC = 0x3, // Carry Clear (or LO)
	LO = 0x3,
	MI = 0x4, // MInus
	PL = 0x5, // PLus
	VS = 0x6, // oVerflow Set
	VC = 0x7, // oVerflow Clear
	HI = 0x8, // HIgher
	LS = 0x9, // Lower or Same
	GE = 0xA, // Greater or Equal
	LT = 0xB, // Less Than
	GT = 0xC, // Greater Than
	LE = 0xD, // Less or Equal
	AL = 0xE, // ALways

	// Note that condition code NV is unpredictable on ARMv3 and ARMv4, and has
	// special meaning for ARMv5 onwards. As such, it should never be used in
	// an instruction encoding unless the special (ARMv5+) meaning is required.
	NV = 0xF  // NeVer
} ConditionCode;

#define firstreg()      R0

verbose_only( extern const char* regNames[]; )
verbose_only( extern const char* condNames[]; )
verbose_only( extern const char* shiftNames[]; )

// abstract to platform specific calls
#define nExtractPlatformFlags(x)    0

#define DECLARE_PLATFORM_STATS()

#define DECLARE_PLATFORM_REGALLOC()                                     \
	Register getAvailableReg(LIns* ins, Register regClass, RegisterMask allow);\
	Register getSuitableRegFor(Register r, LIns* curins)               \

#if defined(NANOJIT_DEBUG)
# define DECLARE_PLATFORM_ASSEMBLER_DEBUG()                             \
	bool             isOp2Imm(uint32_t literal);                     \
	uint32_t         decOp2Imm(uint32_t enc);
#else
// define stubs, for code that defines NANOJIT_VERBOSE without DEBUG
# define DECLARE_PLATFORM_ASSEMBLER_DEBUG()                             \
	inline bool      isOp2Imm(uint32_t ) { return true; }            \
	inline uint32_t  decOp2Imm(uint32_t ) { return 0; }
#endif

#define DECLARE_PLATFORM_ASSEMBLER()                                                     \
	static const unsigned int _nCodeBlockSizeThresholdInBytes = 2048;                    \
																				         \
	DECLARE_PLATFORM_ASSEMBLER_DEBUG()                                                   \
																				         \
	void            BranchWithLink(NIns* addr);                                          \
	inline void     BLX(Register addr, bool chk = true);                                 \
	void            JMP_far(NIns*);                                                      \
	void            B_cond_chk(ConditionCode, const NIns*, bool);                        \
	bool            hardenNopInsertion(const Config& /*c*/) { return false; }            \
	void            asm_immd_nochk(Register, int32_t, int32_t);                          \
	void            asm_pushstate();                                                     \
	void            asm_popstate();                                                      \
	void            asm_savepc();                                                        \
	void            asm_restorepc();                                                     \
	void            asm_discardpc();                                                     \
	void            asm_cmpi(Register, int32_t imm);                                     \
	void            asm_ldr_chk(Register d, Register b, int32_t off, bool chk);          \
	int32_t         asm_str(Register rt, Register rr, int32_t off);                      \
	void            asm_cmp(LIns *cond);                                                 \
	void            asm_cmpd(LIns *cond);                                                \
	void            asm_ld_imm(Register d, int32_t imm, bool chk = true);                \
	void            asm_add_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_adc_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_sub_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_sbc_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_and_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_bic_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_orr_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_eor_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_rcr_imm(Register rd, Register rn, int32_t imm, int stat = 0);    \
	void            asm_mul(LIns* ins);                                                  \
	bool			encOp2Imm(uint32_t literal, uint32_t * enc);                         \
	inline uint32_t CountLeadingZeroes(uint32_t data);                                   \
                                                                                         \
	void            asm_immf_nochk(Register, int32_t);                                   \
	void            asm_cmpf4(LIns *cond);                                               \
	Register        getAvailableReg(LIns* ins, Register regClass, RegisterMask m)        \
			   { return _allocator.getAvailableReg(ins, regClass, m); }                  \
																						 \
	struct ConstantSlot                                                                  \
	{                                                                                    \
		NIns*		loadNativeIns;                                                       \
		uint32_t	value;                                                               \
	};                                                                                   \
	static const size_t CONSTANT_SLOT_COUNT = 64;                                        \
	unsigned int	_nConstantSlotCount;                                                 \
	ConstantSlot    _nConstantSlot[CONSTANT_SLOT_COUNT];                                 \
																						 \
	NIns*			allocateConstantSlot(NIns* loadIns, uint32_t value);                 \
	void			processPendingConstantSlots();                                       \
	void			nPatchLoadPcImm(NIns* loadIns, NIns* constantSlot);                  \
	bool			nNotifyPageJump();                                                   \
	

const int NANOJIT_SMALLEST_UNDERRUN_PROT = 4;



} // namespace nanojit
#endif // __nanojit_NativeARM__
