

#ifndef __nanojit_NativeARM_Macros_
#define __nanojit_NativeARM_Macros_


namespace nanojit
{

#define IMM32(imm)  *(--_nIns) = (NIns)((imm));

#define OP_IMM  (1<<25)
#define OP_STAT (1<<20)

#define COND_AL ((uint32_t)AL<<28)

typedef enum {
	LSL_imm = 0, // LSL #c - Logical Shift Left
	LSL_reg = 1, // LSL Rc - Logical Shift Left
	LSR_imm = 2, // LSR #c - Logical Shift Right
	LSR_reg = 3, // LSR Rc - Logical Shift Right
	ASR_imm = 4, // ASR #c - Arithmetic Shift Right
	ASR_reg = 5, // ASR Rc - Arithmetic Shift Right
	ROR_imm = 6, // Rotate Right (c != 0)
	RRX     = 6, // Rotate Right one bit with extend (c == 0)
	ROR_reg = 7  // Rotate Right
} ShiftOperator;
#define IsShift(sh)    (((sh) >= LSL_imm) && ((sh) <= ROR_reg))

#define LD32_size 8

#define BEGIN_NATIVE_CODE(x)                    \
	{ DWORD* _nIns = (uint8_t*)x

#define END_NATIVE_CODE(x)                      \
	(x) = (dictwordp*)_nIns; }

// BX
#define BX(_r)  do {                                                    \
		underrunProtect(4);                                             \
		NanoAssert(IsGpReg(_r));                                        \
		*(--_nIns) = (NIns)( COND_AL | (0x12<<20) | (0xFFF<<8) | (1<<4) | (REGNUM(_r))); \
		asm_output("bx %s", gpn(_r)); } while(0)

/*
 * ALU operations
 */

enum {
	ARM_and = 0,
	ARM_eor = 1,
	ARM_sub = 2,
	ARM_rsb = 3,
	ARM_add = 4,
	ARM_adc = 5,
	ARM_sbc = 6,
	ARM_rsc = 7,
	ARM_tst = 8,
	ARM_teq = 9,
	ARM_cmp = 10,
	ARM_cmn = 11,
	ARM_orr = 12,
	ARM_mov = 13,
	ARM_bic = 14,
	ARM_mvn = 15
};



#define IsCond(cc)        (((cc) >= EQ) && ((cc) <= AL))

// Bit 0 of the condition code can be flipped to obtain the opposite condition.
// However, this won't work for AL because its opposite  NV  has special
// meaning.
#define OppositeCond(cc)  ((ConditionCode)((unsigned int)(cc)^0x1))

static inline bool IsFpSReg(Register _r) { return REGNUM(_r) >= REGNUM(FirstSFloatReg) && REGNUM(_r) < REGNUM(FirstDFloatReg); }
static inline bool IsFpDReg(Register _r) { return REGNUM(_r) >= REGNUM(FirstDFloatReg) && REGNUM(_r) < REGNUM(FirstQFloatReg); }
static inline bool IsFpQReg(Register _r) { return REGNUM(_r) >= REGNUM(FirstQFloatReg) && REGNUM(_r) <= REGNUM(Q15); }
static inline bool IsGpReg(Register _r) { return REGNUM(_r) < REGNUM(FirstSFloatReg); }
#define FpSRegNum(_fpr)  (REGNUM(_fpr) - REGNUM(FirstSFloatReg))
#define FpDRegNum(_fpr)  ((REGNUM(_fpr) - REGNUM(FirstDFloatReg)) ^ 16) // mind the reversed order of the banks
#define FpQRegNum(_fpr)  (REGNUM(_fpr) - REGNUM(FirstQFloatReg))
static inline Register SReg(uint32_t num)  { Register r = { REGNUM(FirstSFloatReg) + (num) }; return r; }
static inline Register DReg(uint32_t num)  { Register r = { REGNUM(FirstDFloatReg) + ((num) ^ 16) }; return r; }  // mind the reversed order of the register banks
static inline Register QReg(uint32_t num)  { Register r = { REGNUM(FirstQFloatReg) + (num)  }; return r; }

#define IsOp(op)      (((ARM_##op) >= ARM_and) && ((ARM_##op) <= ARM_mvn))

// ALU operation with register and 8-bit immediate arguments
//  S   - bit, 0 or 1, whether the CPSR register is updated
//  rd  - destination register
//  rl  - first (left) operand register
//  op2imm  - operand 2 immediate. Use encOp2Imm (from NativeARM.cpp) to calculate this.
#define ALUi(cond, op, S, rd, rl, op2imm)   ALUi_chk(cond, op, S, rd, rl, op2imm, 1)
#define ALUi_chk(cond, op, S, rd, rl, op2imm, chk) do {\
		if (chk) underrunProtect(4);\
		NanoAssert(IsCond(cond));\
		NanoAssert(IsOp(op));\
		NanoAssert(((S)==0) || ((S)==1));\
		NanoAssert(IsGpReg(rd) && IsGpReg(rl));\
		NanoAssert(isOp2Imm(op2imm));\
		*(--_nIns) = (NIns) (((unsigned int)cond)<<28 | OP_IMM | (ARM_##op)<<21 | (S)<<20 | (REGNUM(rl))<<16 | (REGNUM(rd))<<12 | (op2imm));\
		if (ARM_##op == ARM_mov || ARM_##op == ARM_mvn) {               \
			asm_output("%s%s%s %s, #0x%X", #op, condNames[cond], (S)?"s":"", gpn(rd), decOp2Imm(op2imm));\
		} else if (ARM_##op >= ARM_tst && ARM_##op <= ARM_cmn) {         \
			NanoAssert(S==1);\
			asm_output("%s%s %s, #0x%X", #op, condNames[cond], gpn(rl), decOp2Imm(op2imm));\
		} else {                                                        \
			asm_output("%s%s%s %s, %s, #0x%X", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rl), decOp2Imm(op2imm));\
		}\
	} while (0)

// ALU operation with two register arguments
//  S   - bit, 0 or 1, whether the CPSR register is updated
//  rd  - destination register
//  rl  - first (left) operand register
//  rr  - first (left) operand register
#define ALUr(cond, op, S, rd, rl, rr)   ALUr_chk(cond, op, S, rd, rl, rr, 1)
#define ALUr_chk(cond, op, S, rd, rl, rr, chk) do {\
		if (chk) underrunProtect(4);\
		NanoAssert(IsCond(cond));\
		NanoAssert(IsOp(op));\
		NanoAssert(((S)==0) || ((S)==1));\
		NanoAssert(IsGpReg(rd) && IsGpReg(rl) && IsGpReg(rr));\
		*(--_nIns) = (NIns) (((unsigned int)cond)<<28 |(ARM_##op)<<21 | (S)<<20 | (REGNUM(rl))<<16 | (REGNUM(rd))<<12 | (REGNUM(rr)));\
		if (ARM_##op == ARM_mov || ARM_##op == ARM_mvn) {               \
			asm_output("%s%s%s %s, %s", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rr));\
		} else if (ARM_##op >= ARM_tst && ARM_##op <= ARM_cmn) {         \
			NanoAssert(S==1);\
			asm_output("%s%s  %s, %s", #op, condNames[cond], gpn(rl), gpn(rr));\
		} else {                                                        \
			asm_output("%s%s%s %s, %s, %s", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rl), gpn(rr));\
		}\
	} while (0)

// ALU operation with two register arguments, with rr operated on by a shift and shift immediate
//  S   - bit, 0 or 1, whether the CPSR register is updated
//  rd  - destination register
//  rl  - first (left) operand register
//  rr  - second (right) operand register
//  sh  - a ShiftOperator
//  imm - immediate argument to shift operator, 5 bits (0..31)
#define ALUr_shi(cond, op, S, rd, rl, rr, sh, imm) do {\
		underrunProtect(4);\
		NanoAssert(IsCond(cond));\
		NanoAssert(IsOp(op));\
		NanoAssert(((S)==0) || ((S)==1));\
		NanoAssert(IsGpReg(rd) && IsGpReg(rl) && IsGpReg(rr));\
		NanoAssert(IsShift(sh));\
		NanoAssert((imm)>=0 && (imm)<32);\
		*(--_nIns) = (NIns) (((unsigned int)cond)<<28 |(ARM_##op)<<21 | (S)<<20 | (REGNUM(rl))<<16 | (REGNUM(rd))<<12 | (imm)<<7 | (sh)<<4 | (REGNUM(rr)));\
		if (ARM_##op == ARM_mov || ARM_##op == ARM_mvn) {               \
			NanoAssert(REGNUM(rl)==0);                                          \
			asm_output("%s%s%s %s, %s, %s #%d", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rr), shiftNames[sh], (imm));\
		} else if (ARM_##op >= ARM_tst && ARM_##op <= ARM_cmn) {         \
			NanoAssert(S==1);\
			NanoAssert(REGNUM(rd)==0);\
			asm_output("%s%s  %s, %s, %s #%d", #op, condNames[cond], gpn(rl), gpn(rr), shiftNames[sh], (imm));\
		} else {                                                        \
			asm_output("%s%s%s %s, %s, %s, %s #%d", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rl), gpn(rr), shiftNames[sh], (imm));\
		}\
	} while (0)

// ALU operation with two register arguments, with rr operated on by a shift and shift register
//  S   - bit, 0 or 1, whether the CPSR register is updated
//  rd  - destination register
//  rl  - first (left) operand register
//  rr  - first (left) operand register
//  sh  - a ShiftOperator
//  rs  - shift operand register
#define ALUr_shr(cond, op, S, rd, rl, rr, sh, rs) do {\
		underrunProtect(4);\
		NanoAssert(IsCond(cond));\
		NanoAssert(IsOp(op));\
		NanoAssert(((S)==0) || ((S)==1));\
		NanoAssert(IsGpReg(rd) && IsGpReg(rl) && IsGpReg(rr) && IsGpReg(rs));\
		NanoAssert(IsShift(sh));\
		*(--_nIns) = (NIns) (((unsigned int)cond)<<28 |(ARM_##op)<<21 | (S)<<20 | (REGNUM(rl))<<16 | (REGNUM(rd))<<12 | (REGNUM(rs))<<8 | (sh)<<4 | (REGNUM(rr)));\
		if (ARM_##op == ARM_mov || ARM_##op == ARM_mvn) {               \
			asm_output("%s%s%s %s, %s, %s %s", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rr), shiftNames[sh], gpn(rs));\
		} else if (ARM_##op >= ARM_tst && ARM_##op <= ARM_cmn) {         \
			NanoAssert(S==1);\
			asm_output("%s%s  %s, %s, %s %s", #op, condNames[cond], gpn(rl), gpn(rr), shiftNames[sh], gpn(rs));\
		} else {                                                        \
			asm_output("%s%s%s %s, %s, %s, %s %s", #op, condNames[cond], (S)?"s":"", gpn(rd), gpn(rl), gpn(rr), shiftNames[sh], gpn(rs));\
		}\
	} while (0)

// --------
// Basic arithmetic operations.
// --------
// Argument naming conventions for these macros:
//  _d      Destination register.
//  _l      First (left) operand.
//  _r      Second (right) operand.
//  _op2imm An operand 2 immediate value. Use encOp2Imm to calculate this.
//  _s      Set to 1 to update the status flags (for subsequent conditional
//          tests). Otherwise, set to 0.

// _d = _l & op
#define BIC(_d, _l, _r, _s)						ALUr(AL, bic, _s, _d, _l, _r);
#define BICi(_d,_l,_op2imm,_s)					ALUi(AL, bic, _s, _d, _l, _op2imm)
#define BICshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, bic, _s, _d, _l, _r, _shift, _rs)
#define BICshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, bic, _s, _d, _l, _r, _shift, _imm)

// _d = _l & op
#define AND(_d, _l, _r, _s)						ALUr(AL, and, _s, _d, _l, _r);
#define ANDi(_d,_l,_op2imm,_s)					ALUi(AL, and, _s, _d, _l, _op2imm)
#define ANDshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, and, _s, _d, _l,_r,  _shift, _rs)
#define ANDshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, and, _s, _d, _l, _r, _shift, _imm)

// _d = _l | op
#define ORR(_d, _l, _r, _s)						ALUr(AL, orr, _s, _d, _l, _r);
#define ORRi(_d,_l,_op2imm,_s)					ALUi(AL, orr, _s, _d, _l, _op2imm)
#define ORRshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, orr, _s, _d, _l, _r, _shift, _rs)
#define ORRshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, orr, _s, _d, _l, _r, _shift, _imm)

// _d = _l ^ op
#define EOR(_d, _l, _r, _s)						ALUr(AL, eor, _s, _d, _l, _r);
#define EORi(_d,_l,_op2imm,_s)					ALUi(AL, eor, _s, _d, _l, _op2imm)
#define EORshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, eor, _s, _d, _l, _r, _shift, _rs)
#define EORshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, eor, _s, _d, _l, _r, _shift, _imm)

// _d = _l + _r
#define ADD(_d,_l,_r, _s)						ALUr(AL, add,  _s, _d, _l, _r)
#define ADDi(_d,_l,_op2imm, _s)					ALUi(AL, add,  _s, _d, _l, _op2imm)
#define ADDshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, add, _s, _d, _l, _r, _shift, _rs)
#define ADDshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, add, _s, _d, _l, _r, _shift, _imm)

// _d = _l + _r + carry
#define ADC(_d,_l,_r,_s)						ALUr(AL, adc, _s, _d, _l, _r)
#define ADCi(_d,_l,_op2imm, _s)					ALUi(AL, adc,  _s, _d, _l, _op2imm)
#define ADCshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, adc, _s, _d, _l, _r, _shift, _rs)
#define ADCshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, adc, _s, _d, _l, _r, _shift, _imm)

// _d = _l - _r
#define SUB(_d,_l,_r,_s)						ALUr(AL, sub, _s, _d, _l, _r)
#define SUBi(_d,_l,_op2imm,_s)					ALUi(AL, sub, _s, _d, _l, _op2imm)
#define SUBshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, sub, _s, _d, _l, _r, _shift, _rs)
#define SUBshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, sub, _s, _d, _l, _r, _shift, _imm)

// _d = _r - _l
#define RSB(_d,_l,_r,_s)						ALUr(AL, rsb, _s, _d, _l, _r)
#define RSBi(_d,_l,_op2imm, _s)					ALUi(AL, rsb,  _s, _d, _l, _op2imm)
#define RSBshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, rsb, _s, _d, _l, _r, _shift, _rs)
#define RSBshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, rsb, _s, _d, _l, _r, _shift, _imm)

// _d = _l - _r - NOT(carry)
#define SBC(_d,_l,_r,_s)						ALUr(AL, sbc, _s, _d, _l, _r)
#define SBCi(_d,_l,_op2imm, _s)					ALUi(AL, sbc,  _s, _d, _l, _op2imm)
#define SBCshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, sbc, _s, _d, _l, _r, _shift, _rs)
#define SBCshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, sbc, _s, _d, _l, _r, _shift, _imm)

// _d = _r - _l - NOT(carry)
#define RSC(_d,_l,_r,_s)						ALUr(AL, rsc, _s, _d, _l, _r)
#define RSCi(_d,_l,_op2imm, _s)					ALUi(AL, rsc,  _s, _d, _l, _op2imm)
#define RSCshr(_d, _l, _r, _shift, _rs, _s)		ALUr_shr(AL, rsc, _s, _d, _l, _r, _shift, _rs)
#define RSCshi(_d, _l, _r, _shift, _imm, _s)	ALUr_shi(AL, rsc, _s, _d, _l, _r, _shift, _imm)



// --------
// Other operations.
// --------

#define MULL(_op_txt, _op, _d, _d_hi, _l, _r, _s)  do {                                                   \
	NanoAssert(ARM_ARCH_AT_LEAST(6) || ((_d   ) != (_l)));                                        \
	NanoAssert(ARM_ARCH_AT_LEAST(6) || ((_d_hi) != (_l)));                                        \
	NanoAssert(IsGpReg(_d) && IsGpReg(_d_hi) && IsGpReg(_l) && IsGpReg(_r));                      \
	NanoAssert(((_d) != PC) && ((_d_hi) != PC) && ((_l) != PC) && ((_r) != PC));                  \
	*(--_nIns) = (NIns)( COND_AL | _op | (_s) << 20 | (REGNUM(_d_hi))<<16 | (REGNUM(_d))<<12 | (REGNUM(_r))<<8 | (REGNUM(_l)) ); \
	asm_output("%s %s, %s, %s, %s",_op_txt,gpn(_d),gpn(_d_hi),gpn(_l),gpn(_r));                       \
} while(0)

// [_d_hi,_d] = _l * _r
#define SMULLs(_d, _d_hi, _l, _r, _s) MULL("SMULLs", 0x00c00090, _d, _d_hi, _l, _r, _s);
#define SMULL(_d, _d_hi, _l, _r) MULL("SMULL", 0x00c00090, _d, _d_hi, _l, _r, 0);
#define UMULLs(_d, _d_hi, _l, _r, _s) MULL("UMULLs", 0x00800090, _d, _d_hi, _l, _r, _s);
#define UMULL(_d, _d_hi, _l, _r) MULL("UMULLs", 0x00800090, _d, _d_hi, _l, _r, 0);

// [_d_hi,_d] += _l * _r
#define SMLALs(_d, _d_hi, _l, _r, _s) MULL("SMLALs", 0x00e00090, _d, _d_hi, _l, _r, _s);
#define SMLAL(_d, _d_hi, _l, _r) MULL("SMLAL", 0x00e00090, _d, _d_hi, _l, _r, 0);
#define UMLALs(_d, _d_hi, _l, _r, _s) MULL("UMLALs", 0x00a00090, _d, _d_hi, _l, _r, _s);
#define UMLAL(_d, _d_hi, _l, _r) MULL("UMLAL", 0x00a00090, _d, _d_hi, _l, _r, 0);

// _d = _l * _r
#define MUL(_d, _l, _r, _stat)  do {                                               \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_ARCH_AT_LEAST(6) || ((_d) != (_l)));              \
		NanoAssert(IsGpReg(_d) && IsGpReg(_l) && IsGpReg(_r));              \
		NanoAssert(((_d) != PC) && ((_l) != PC) && ((_r) != PC));           \
		*(--_nIns) = (NIns)( COND_AL | (_stat)<<20 | (REGNUM(_d))<<16 | (REGNUM(_r))<<8 | 0x90 | (REGNUM(_l)) );  \
		asm_output("mul %s, %s, %s",gpn(_d),gpn(_l),gpn(_r));               \
} while(0)

// RSBS _d, _r
// _d = 0 - _r
#define RSBS(_d,_r, _stat) ALUi(AL, rsb, _stat, _d, _r, 0)

// MVN
// _d = ~_r (one's compliment)
#define MVN(_d,_r,_stat)                          ALUr(AL, mvn, _stat, _d, SBZ, _r)
#define MVNis_chk(_d,_op2imm,_stat,_chk)    ALUi_chk(AL, mvn, _stat, _d, R0, op2imm, _chk)
#define MVNis(_d,_op2imm,_stat)             MVNis_chk(_d,_op2imm,_stat,1);

// Logical Shift Right (LSR) rotates the bits without maintaining sign extensions.
// MOVS _d, _r, LSR <_s>
// _d = _r >> _s
#define LSR(_d, _l, _r,_s) ALUr_shr(AL, mov, _s, _d, SBZ, _l, LSR_reg, _r)

// Logical Shift Right (LSR) rotates the bits without maintaining sign extensions.
// MOVS _d, _r, LSR #(_imm & 0x1f)
// _d = _r >> (_imm & 0x1f)
#define LSRi(_d, _r, _imm, _s)  ALUr_shi(AL, mov, _s, _d, SBZ, _r, LSR_imm, (_imm & 0x1f))

// Arithmetic Shift Right (ASR) maintains the sign extension.
// MOVS _d, _r, ASR <_s>
// _d = _r >> _s
#define ASR( _d, _l, _r, _s) ALUr_shr(AL, mov, _s, _d, SBZ, _l, ASR_reg, _r)

// Arithmetic Shift Right (ASR) maintains the sign extension.
// MOVS _r, _r, ASR #(_imm & 0x1f)
// _d = _r >> (_imm & 0x1f)
#define ASRi(_d, _r, _imm, _s) ALUr_shi(AL, mov, 1, _d, SBZ, _r, ASR_imm, (_imm & 0x1f))

// Logical Shift Left (LSL).
// MOVS _d, _r, LSL <_s>
// _d = _r << _s
#define LSL(_d, _l, _r, _s) ALUr_shr(AL, mov, _s, _d, SBZ, _l, LSL_reg, _r)

// Logical Shift Left (LSL).
// MOVS _d, _r, LSL #(_imm & 0x1f)
// _d = _r << (_imm & 0x1f)
#define LSLi(_d, _r, _imm, _s) ALUr_shi(AL, mov, _s, _d, SBZ, _r, LSL_imm, (_imm & 0x1f))


// Rotate right (ROR).
// MOVS _d, _r, ROR <_s>
#define ROR(_d, _l, _r, _s) ALUr_shr(AL, mov, _s, _d, SBZ, _l, ROR_reg, _r)

// Rotate right immediate (ROR).
// MOVS _d, _r, ROR <#imm & 0x1f>
#define RORi(_d, _r, _imm, _s) ALUr_shi(AL, mov, _s, _d, SBZ, _r, ROR_imm, (_imm & 0x1f))

// Rotate right extended (with carry)
#define RRX(_d, _r, _s) ALUr_shi(AL, mov, _s, _d, SBZ, _r, RRX, 0)

// TST
#define TST(_l,_r)      ALUr(AL, tst, 1, SBZ, _l, _r)
#define TSTi(_d,_imm)   ALUi(AL, tst, 1, SBZ, _d, _imm)

// CMP
#define CMP(_l,_r)  ALUr(AL, cmp, 1, SBZ, _l, _r)
#define CMN(_l,_r)  ALUr(AL, cmn, 1, SBZ, _l, _r)

// MOV
#define MOVis_chk(_d,_op2imm,_stat,_chk)    ALUi_chk(AL, mov, _stat, _d, R0, op2imm, _chk)
#define MOVis(_d,_op2imm,_stat)             MOVis_chk(_d,_op2imm,_stat,1)
#define MOVi(_d,_op2imm)                    MOVis(_d,_op2imm,0);

#define MOV_cond(_cond,_d,_s)               ALUr(_cond, mov, 0, _d, R0, _s)

#define MOV(dr,sr)   MOV_cond(AL, dr, sr)

// _d = [_b+off]
#define LDR(_d,_b,_off)        asm_ldr_chk(_d,_b,_off,1)
#define LDR_nochk(_d,_b,_off)  asm_ldr_chk(_d,_b,_off,0)

// _d = [_b + _x<<_s]
#define LDR_scaled(_d, _b, _x, _s) do { \
		NanoAssert(((_s)&31) == _s);\
		NanoAssert(IsGpReg(_d) && IsGpReg(_b) && IsGpReg(_x));\
		underrunProtect(4);\
		*(--_nIns) = (NIns)(COND_AL | (0x79<<20) | ((_b)<<16) | ((_d)<<12) | ((_s)<<7) | (_x));\
		asm_output("ldr %s, [%s, +%s, LSL #%d]", gpn(_d), gpn(_b), gpn(_x), (_s));\
	} while (0)

// _d = #_imm
#define LDi(_d,_imm) asm_ld_imm(_d,_imm)

// MOVW and MOVT are ARMv6T2 or newer only

// MOVW -- writes _imm into _d, zero-extends.
#define MOVWi_cond_chk(_cond,_d,_imm,_chk) do {                         \
		NanoAssert(isU16(_imm));                                        \
		NanoAssert(IsGpReg(_d));                                        \
		NanoAssert(IsCond(_cond));                                      \
		if (_chk) underrunProtect(4);                                   \
		*(--_nIns) = (NIns)( ((unsigned int)_cond)<<28 | 3<<24 | 0<<20 | (((_imm)>>12)&0xf)<<16 | (REGNUM(_d))<<12 | ((_imm)&0xfff) ); \
		asm_output("movw%s %s, #0x%x", condNames[_cond], gpn(_d), (_imm)); \
	} while (0)

#define MOVWi(_d,_imm)              MOVWi_cond_chk(AL, _d, _imm, 1)
#define MOVWi_chk(_d,_imm,_chk)     MOVWi_cond_chk(AL, _d, _imm, _chk)
#define MOVWi_cond(_cond,_d,_imm)   MOVWi_cond_chk(_cond, _d, _imm, 1)

// MOVT -- writes _imm into top halfword of _d, does not affect bottom halfword
#define MOVTi_cond_chk(_cond,_d,_imm,_chk) do {                         \
		NanoAssert(isU16(_imm));                                        \
		NanoAssert(IsGpReg(_d));                                        \
		NanoAssert(IsCond(_cond));                                      \
		if (_chk) underrunProtect(4);                                   \
		*(--_nIns) = (NIns)( ((unsigned int)_cond)<<28 | 3<<24 | 4<<20 | (((_imm)>>12)&0xf)<<16 | (REGNUM(_d))<<12 | ((_imm)&0xfff) ); \
		asm_output("movt%s %s, #0x%x", condNames[_cond], gpn(_d), (_imm)); \
	} while (0)

#define MOVTi(_d,_imm)              MOVTi_cond_chk(AL, _d, _imm, 1)
#define MOVTi_chk(_d,_imm,_chk)     MOVTi_cond_chk(AL, _d, _imm, _chk)
#define MOVTi_cond(_cond,_d,_imm)   MOVTi_cond_chk(_cond, _d, _imm, 1)

// Load a byte (8 bits). The offset range is 4095.
#define LDRB(_d,_n,_off) do {                                           \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		underrunProtect(4);                                             \
		if (_off < 0) {                                                 \
			NanoAssert(isU12(-(_off)));                                 \
			*(--_nIns) = (NIns)( COND_AL | (0x55<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((-(_off))&0xfff)  ); \
		} else {                                                        \
			NanoAssert(isU12(_off));                                    \
			*(--_nIns) = (NIns)( COND_AL | (0x5D<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((_off)&0xfff)  ); \
		}                                                               \
		asm_output("ldrb %s, [%s,#%d]", gpn(_d),gpn(_n),(_off));        \
	} while(0)

// Load a byte (8 bits), sign-extend to 32 bits. The offset range is
// 255 (different from LDRB, same as LDRH/LDRSH)
#define LDRSB(_d,_n,_off) do {                                          \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		underrunProtect(4);                                             \
		if (_off < 0) {                                                 \
			NanoAssert(isU8(-(_off)));                                  \
			*(--_nIns) = (NIns)( COND_AL | (0x15<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xD)<<4) | (((-(_off))&0xf0)<<4) | ((-(_off))&0xf) ); \
		} else {                                                        \
			NanoAssert(isU8(_off));                                     \
			*(--_nIns) = (NIns)( COND_AL | (0x1D<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xD)<<4) | (((_off)&0xf0)<<4) | ((_off)&0xf) ); \
		}                                                               \
		asm_output("ldrsb %s, [%s,#%d]", gpn(_d),gpn(_n),(_off));       \
	} while(0)

// Load and sign-extend a half word (16 bits). The offset range is 255, and
// must be aligned to two bytes on some architectures (the caller is responsible
// for ensuring appropriate alignment)
#define LDRH(_d,_n,_off) do {                                           \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		underrunProtect(4);                                             \
		if (_off < 0) {                                                 \
			NanoAssert(isU8(-(_off)));                                  \
			*(--_nIns) = (NIns)( COND_AL | (0x15<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xB)<<4) | (((-(_off))&0xf0)<<4) | ((-(_off))&0xf) ); \
		} else {                                                        \
			NanoAssert(isU8(_off));                                     \
			*(--_nIns) = (NIns)( COND_AL | (0x1D<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xB)<<4) | (((_off)&0xf0)<<4) | ((_off)&0xf) ); \
		}                                                               \
		asm_output("ldrh %s, [%s,#%d]", gpn(_d),gpn(_n),(_off));       \
	} while(0)

// Load and sign-extend a half word (16 bits). The offset range is 255, and
// must be aligned to two bytes on some architectures (the caller is responsible
// for ensuring appropriate alignment)
#define LDRSH(_d,_n,_off) do {                                          \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		underrunProtect(4);                                             \
		if (_off < 0) {                                                 \
			NanoAssert(isU8(-(_off)));                                  \
			*(--_nIns) = (NIns)( COND_AL | (0x15<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xF)<<4) | (((-(_off))&0xf0)<<4) | ((-(_off))&0xf) ); \
		} else {                                                        \
			NanoAssert(isU8(_off));                                     \
			*(--_nIns) = (NIns)( COND_AL | (0x1D<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((0xF)<<4) | (((_off)&0xf0)<<4) | ((_off)&0xf) ); \
		}                                                               \
		asm_output("ldrsh %s, [%s,#%d]", gpn(_d),gpn(_n),(_off));       \
	} while(0)

// Valid offset for STR and STRB is +/- 4095, STRH only has +/- 255
#define STR(_d,_n,_off) do {                                            \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		NanoAssert(isU12(_off) || isU12(-(_off)));                      \
		underrunProtect(4);                                             \
		if ((_off)<0)   *(--_nIns) = (NIns)( COND_AL | (0x50<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((-(_off))&0xFFF) ); \
		else            *(--_nIns) = (NIns)( COND_AL | (0x58<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((_off)&0xFFF) ); \
		asm_output("str %s, [%s, #%d]", gpn(_d), gpn(_n), (_off)); \
	} while(0)

#define STRB(_d,_n,_off) do {                                           \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		NanoAssert(isU12(_off) || isU12(-(_off)));                      \
		underrunProtect(4);                                             \
		if ((_off)<0)   *(--_nIns) = (NIns)( COND_AL | (0x54<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((-(_off))&0xFFF) ); \
		else            *(--_nIns) = (NIns)( COND_AL | (0x5C<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | ((_off)&0xFFF) ); \
		asm_output("strb %s, [%s, #%d]", gpn(_d), gpn(_n), (_off)); \
	} while(0)

// Only +/- 255 range, unlike STRB/STR
#define STRH(_d,_n,_off) do {                                           \
		NanoAssert(IsGpReg(_d) && IsGpReg(_n));                         \
		underrunProtect(4);                                             \
		if ((_off)<0) {                                                 \
			NanoAssert(isU8(-(_off)));                                  \
			*(--_nIns) = (NIns)( COND_AL | (0x14<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | (((-(_off))&0xF0)<<4) | (0xB<<4) | ((-(_off))&0xF) ); \
		} else {                                                        \
			NanoAssert(isU8(_off));                                     \
			*(--_nIns) = (NIns)( COND_AL | (0x1C<<20) | ((REGNUM(_n))<<16) | ((REGNUM(_d))<<12) | (((_off)&0xF0)<<4) | (0xB<<4) | ((_off)&0xF) ); \
		}                                                               \
		asm_output("strh %s, [%s, #%d]", gpn(_d), gpn(_n), (_off));     \
	} while(0)

// Encode a breakpoint. The ID is not important and is ignored by the
// processor, but it can be useful as a marker when debugging emitted code.
#define BKPT_insn       ((NIns)( COND_AL | (0x12<<20) | (0x7<<4) ))
#define BKPTi_insn(id)  ((NIns)(BKPT_insn | ((id << 4) & 0xfff00) | (id & 0xf)));

#define BKPT_nochk()    BKPTi_nochk(0)
#define BKPTi_nochk(id) do {                                \
		NanoAssert((id & 0xffff) == id);                    \
		*(--_nIns) = BKPTi_insn(id);                        \
		} while (0)

// STMFD SP!, {reg}
#define PUSHr(_r)  do {                                                 \
		underrunProtect(4);                                             \
		NanoAssert(IsGpReg(_r));                                        \
		*(--_nIns) = (NIns)( COND_AL | (0x92<<20) | (REGNUM(SP)<<16) | rmask(_r) ); \
		asm_output("push %s",gpn(_r)); } while (0)

// STMFD SP!,{reglist}
#define PUSH_mask(_mask)  do {                                          \
		underrunProtect(4);                                             \
		NanoAssert(isU16(_mask));                                       \
		*(--_nIns) = (NIns)( COND_AL | (0x92<<20) | (REGNUM(SP)<<16) | (_mask) ); \
		asm_output("push %x", static_cast<unsigned int>(_mask));} while (0)

// LDMFD SP!,{reg}
#define POPr(_r) do {                                                   \
		underrunProtect(4);                                             \
		NanoAssert(IsGpReg(_r));                                        \
		*(--_nIns) = (NIns)( COND_AL | (0x8B<<20) | (REGNUM(SP)<<16) | rmask(_r) ); \
		asm_output("pop %s",gpn(_r));} while (0)

// LDMFD SP!,{reglist}
#define POP_mask(_mask) do {                                            \
		underrunProtect(4);                                             \
		NanoAssert(isU16(_mask));                                       \
		*(--_nIns) = (NIns)( COND_AL | (0x8B<<20) | (REGNUM(SP)<<16) | (_mask) ); \
		asm_output("pop %x", static_cast<unsigned int>(_mask));} while (0)

/*
"   0x00234c8c <+84>:\tff 1f 2d e9\tpush\t{r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, r10, r11, r12}\n"
"   0x00234c90 <+88>:\t01 b0 a0 e1\tmov\tr11, r1\n"
"   0x00234c94 <+92>:\t30 ff 2f e1\tblx\tr0\n"
"   0x00234c98 <+96>:\t00 00 a0 e1\tnop\t\t\t; (mov r0, r0)\n"
"   0x00234c9c <+100>:\tff 1f bd e8\tpop\t{r0, r1, r2, r3, r4, r5, r6, r7, r8, r9, r10, r11, r12}\n"
*/



// PC always points to current instruction + 8, so when calculating pc-relative
// offsets, use PC+8.
#define PC_OFFSET_FROM(target,frompc) ((intptr_t)(target) - ((intptr_t)(frompc) + 8))

#define B_cond(_c,_t)                           \
	B_cond_chk(_c,_t,1)

#define B_nochk(_t)                             \
	B_cond_chk(AL,_t,0)

#define B(t)    B_cond(AL,t)
#define BHI(t)  B_cond(HI,t)
#define BLS(t)  B_cond(LS,t)
#define BHS(t)  B_cond(HS,t)
#define BLO(t)  B_cond(LO,t)
#define BEQ(t)  B_cond(EQ,t)
#define BNE(t)  B_cond(NE,t)
#define BLT(t)  B_cond(LT,t)
#define BGE(t)  B_cond(GE,t)
#define BLE(t)  B_cond(LE,t)
#define BGT(t)  B_cond(GT,t)
#define BVS(t)  B_cond(VS,t)
#define BVC(t)  B_cond(VC,t)
#define BCC(t)  B_cond(CC,t)
#define BCS(t)  B_cond(CS,t)

#define JMP(t) B(t)
#define JMP_nochk(t) B_nochk(t)

// MOV(cond) _r, #1
// MOV(!cond) _r, #0
#define SET(_r,_cond) do {                                              \
	ConditionCode _opp = OppositeCond(_cond);                           \
	underrunProtect(8);                                                 \
	*(--_nIns) = (NIns)( ((unsigned int)_opp<<28) | (0x3A<<20) | ((REGNUM(_r))<<12) | (0) ); \
	*(--_nIns) = (NIns)( ((unsigned int)_cond<<28) | (0x3A<<20) | ((REGNUM(_r))<<12) | (1) ); \
	asm_output("mov%s %s, #1", condNames[_cond], gpn(_r));              \
	asm_output("mov%s %s, #0", condNames[_opp], gpn(_r));               \
	} while (0)

#define SETEQ(r)    SET(r,EQ)
#define SETNE(r)    SET(r,NE)
#define SETLT(r)    SET(r,LT)
#define SETLE(r)    SET(r,LE)
#define SETGT(r)    SET(r,GT)
#define SETGE(r)    SET(r,GE)
#define SETLO(r)    SET(r,LO)
#define SETLS(r)    SET(r,LS)
#define SETHI(r)    SET(r,HI)
#define SETHS(r)    SET(r,HS)
#define SETCS(r)    SET(r,CS)
#define SETCC(r)    SET(r,CC)
#define SETMI(r)    SET(r,MI)
#define SETPL(r)    SET(r,PL)
#define SETVS(r)    SET(r,VS)
#define SETVC(r)    SET(r,VC)

// Load and sign extend a 16-bit value into a reg
#define MOVSX(_d,_off,_b) do {                                          \
		if ((_off)>=0) {                                                \
			if ((_off)<256) {                                           \
				underrunProtect(4);                                     \
				*(--_nIns) = (NIns)( COND_AL | (0x1D<<20) | ((_b)<<16) | ((_d)<<12) |  ((((_off)>>4)&0xF)<<8) | (0xF<<4) | ((_off)&0xF)  ); \
			} else if ((_off)<=510) {                                   \
				underrunProtect(8);                                     \
				int rem = (_off) - 255;                                 \
				NanoAssert(rem<256);                                    \
				*(--_nIns) = (NIns)( COND_AL | (0x1D<<20) | ((_d)<<16) | ((_d)<<12) |  ((((rem)>>4)&0xF)<<8) | (0xF<<4) | ((rem)&0xF)  ); \
				*(--_nIns) = (NIns)( COND_AL | OP_IMM | (1<<23) | ((_b)<<16) | ((_d)<<12) | (0xFF) ); \
			} else {                                                    \
				underrunProtect(16);                                    \
				int rem = (_off) & 3;                                   \
				*(--_nIns) = (NIns)( COND_AL | (0x19<<20) | ((_b)<<16) | ((_d)<<12) | (0xF<<4) | (_d) ); \
				asm_output("ldrsh %s,[%s, #%d]",gpn(_d), gpn(_b), (_off)); \
				*(--_nIns) = (NIns)( COND_AL | OP_IMM | (1<<23) | ((_d)<<16) | ((_d)<<12) | rem ); \
				*(--_nIns) = (NIns)( COND_AL | (0x1A<<20) | ((_d)<<12) | (2<<7)| (_d) ); \
				*(--_nIns) = (NIns)( COND_AL | (0x3B<<20) | ((_d)<<12) | (((_off)>>2)&0xFF) ); \
				asm_output("mov %s,%d",gpn(_d),(_off));                \
			}                                                           \
		} else {                                                        \
			if ((_off)>-256) {                                          \
				underrunProtect(4);                                     \
				*(--_nIns) = (NIns)( COND_AL | (0x15<<20) | ((_b)<<16) | ((_d)<<12) |  ((((-(_off))>>4)&0xF)<<8) | (0xF<<4) | ((-(_off))&0xF)  ); \
				asm_output("ldrsh %s,[%s, #%d]",gpn(_d), gpn(_b), (_off)); \
			} else if ((_off)>=-510){                                   \
				underrunProtect(8);                                     \
				int rem = -(_off) - 255;                                \
				NanoAssert(rem<256);                                    \
				*(--_nIns) = (NIns)( COND_AL | (0x15<<20) | ((_d)<<16) | ((_d)<<12) |  ((((rem)>>4)&0xF)<<8) | (0xF<<4) | ((rem)&0xF)  ); \
				*(--_nIns) = (NIns)( COND_AL | OP_IMM | (1<<22) | ((_b)<<16) | ((_d)<<12) | (0xFF) ); \
			} else NanoAssert(0);                                        \
		}                                                               \
	} while(0)

/*
 * VFP, general encoding utilities
 */

#define split_qreg_enc(_Qd,D,Vd)  do{  Vd = FpQRegNum(_Qd);NanoAssert(Vd <= 15);\
									   Vd *= 2;                                 \
									   D = (Vd & 0x10) >> 4;                    \
									   Vd = Vd & 0xf; } while(0)
#define split_dreg_enc(_Dd,D,Vd)  do{  Vd = FpDRegNum(_Dd);NanoAssert(Vd <= 31);\
									   D = (Vd & 0x10) >> 4;                    \
									   Vd = Vd & 0xf; } while(0)
#define split_sreg_enc(_Sd,D,Vd)  do{  Vd = FpSRegNum(_Sd);NanoAssert(Vd <= 31);\
									   D = (Vd & 0x1);                          \
									   Vd = Vd >> 1; } while(0)
	
#define encodeCommon(_Dm,M,Vm,sOk,dOk,qOk)                                      \
		if(IsFpSReg(_Dm)){                                                      \
			NanoAssertMsg( sOk, #_Dm" register is Sn but Sn inacceptable" );\
			split_sreg_enc(_Dm,M,Vm);                                           \
		} else if(IsFpDReg(_Dm)){                                               \
			NanoAssertMsg( dOk, #_Dm" register is Dn but Dn inacceptable" );\
			split_dreg_enc(_Dm,M,Vm);                                           \
		} else if(IsFpQReg(_Dm)){                                               \
			NanoAssertMsg( qOk, #_Dm" register is Qn but Qn inacceptable" );\
			split_qreg_enc(_Dm,M,Vm);                                           \
		} else {                                                                \
			NanoAssertMsg( dOk, #_Dm" register should be FP register" );    \
			M=Vm=0;                                                             \
		}


#define __encodeRegM(_Dm,sOk,dOk,qOk)                               \
		int M,Vm;                                                   \
		encodeCommon(_Dm,M,Vm,sOk,dOk,qOk);                         \
		DmEnc = (M << 5) | Vm;                                      
#define __encodeRegN(_Dn,sOk,dOk,qOk)                               \
		int N,Vn;                                                   \
		encodeCommon(_Dn,N,Vn,sOk,dOk,qOk);                         \
		DnEnc = (N<<7) | (Vn<<16);                                  
#define __encodeRegD(_Dd,sOk,dOk,qOk)                               \
		int D,Vd;                                                   \
		encodeCommon(_Dd,D,Vd,sOk,dOk,qOk);                         \
		DdEnc = (Vd<<12) | (D<<22);                                 
#define DFlag(_Dd) ( IsFpDReg(_Dd) ? (1 << 8) : 0 )
	
//#define encodeRegM(...) __encodeRegM(__VA_ARGS__)        
//#define encodeRegN(...) __encodeRegN(__VA_ARGS__)        
//#define encodeRegD(...) __encodeRegD(__VA_ARGS__)        
#define encodeRegM(a,b,c,d) __encodeRegM(a,b,c,d)        
#define encodeRegN(a,b,c,d) __encodeRegN(a,b,c,d)        
#define encodeRegD(a,b,c,d) __encodeRegD(a,b,c,d)        

//#define SD_  true , true , false
//#define S__  true , false, false
//#define _D_  false, true , false
//#define __Q  false, false, true
	
/*
 * VFP, instructions
 */

#define FMDRR(_Dm,_Rd,_Rn) do {                                         \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(IsGpReg(_Rd) && IsGpReg(_Rn));                       \
		int DmEnc; encodeRegM(_Dm,false, true , false);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xC4<<20) | ((REGNUM(_Rn))<<16) | ((REGNUM(_Rd))<<12) | (0xB1<<4) | DmEnc ); \
		asm_output("fmdrr %s,%s,%s", gpn(_Dm), gpn(_Rd), gpn(_Rn));     \
	} while (0)

#define FMRRD(_Rd,_Rn,_Dm) do {                                         \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(IsGpReg(_Rd) && IsGpReg(_Rn));                       \
		int DmEnc; encodeRegM(_Dm,false, true , false);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xC5<<20) | ((REGNUM(_Rn))<<16) | ((REGNUM(_Rd))<<12) | (0xB1<<4) | DmEnc ); \
		asm_output("fmrrd %s,%s,%s", gpn(_Rd), gpn(_Rn), gpn(_Dm));     \
	} while (0)

#define FMRDH(_Rd,_Dn) do {                                             \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(IsGpReg(_Rd) );                                      \
		int DnEnc; encodeRegN(_Dn,false, true , false);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xE3<<20) | DnEnc | ((REGNUM(_Rd))<<12) | 0xB10 ); \
		asm_output("fmrdh %s,%s", gpn(_Rd), gpn(_Dn));                  \
	} while (0)

#define FMRDL(_Rd,_Dn) do {                                             \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(IsGpReg(_Rd));                                       \
		int DnEnc; encodeRegN(_Dn,DnEnc,false, true , false);                           \
		*(--_nIns) = (NIns)( COND_AL | (0xE1<<20) | DnEnc | ((REGNUM(_Rd))<<12) | 0xB10 ); \
		asm_output("fmrdh %s,%s", gpn(_Rd), gpn(_Dn));                  \
	} while (0)

#define VLDR_chk(_Dd,_Rn,_offs,_chk) do{ \
		if(_chk) underrunProtect(4);                                    \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(((_offs) % 4) == 0);                                 \
		NanoAssert(isU8((_offs)/4) || isU8(-(_offs)/4));                \
		NanoAssert( IsGpReg(_Rn));                                      \
		int negflag = 1<<23;                                            \
		int dflag = DFlag(_Dd);                                         \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                 \
		intptr_t offs = (_offs);                                        \
		if (_offs < 0) {                                                \
			negflag = 0<<23;                                            \
			offs = -(offs);                                             \
		}                                                               \
		*(--_nIns) = (NIns)( COND_AL | (0xD1<<20) | ((REGNUM(_Rn))<<16) | DdEnc | (0xA<<8) | dflag | negflag | ((offs>>2)&0xff) ); \
		asm_output("fld%c %s,%s(%d)", dflag?'d':'s', gpn(_Dd), gpn(_Rn), _offs); \
	} while(0)

#define VLDQR(_Qd,_Rn) do {                                             \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert( IsGpReg(_Rn));                                      \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xC9<<20) | ((REGNUM(_Rn))<<16) | DdEnc | 0xB04 );        \
		asm_output("vldm <d%d,d%d>,%s // i.e. %s",  Vd+D*16,Vd+D*16+1,gpn(_Rn),gpn(_Qd)); \
	} while (0)

#define FLDD_chk VLDR_chk
#define FLDS_chk VLDR_chk
#define FLDD(_Dd,_Rn,_offs) FLDD_chk(_Dd,_Rn,_offs,1)
#define FLDS(_Sd,_Rn,_offs) FLDS_chk(_Sd,_Rn,_offs,1)
#define VLDR FLDD
	
	
#define VSTR(_Dd,_Rn,_offs)  do {                                       \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert(((_offs)%4) == 0);                                   \
		NanoAssert(isU8((_offs)/4) || isU8(-(_offs)/4));                \
		NanoAssert( IsGpReg(_Rn));                                      \
		int dflag = DFlag(_Dd);                                         \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                 \
		int negflag = 1<<23;                                            \
		intptr_t offs = (_offs);                                        \
		if (_offs < 0) {                                                \
			negflag = 0<<23;                                            \
			offs = -(offs);                                             \
		}                                                               \
		*(--_nIns) = (NIns)( COND_AL | (0xD0<<20) | ((REGNUM(_Rn))<<16) | (DdEnc) | (0xA<<8) | dflag | negflag | ((offs>>2)&0xff) ); \
		asm_output("fst%c %s,%s(%d)", dflag?'d':'s',gpn(_Dd), gpn(_Rn), _offs);           \
	} while (0)


#define VSTQR(_Qd,_Rn) do {                                             \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		NanoAssert( IsGpReg(_Rn));                                      \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xC8<<20) | ((REGNUM(_Rn))<<16) | DdEnc | 0xB04 );        \
		asm_output("vstm %s,<d%d,d%d> // i.e. %s", gpn(_Rn), Vd+D*16,Vd+D*16+1,gpn(_Qd)); \
	 } while (0)
#define FSTS  VSTR
#define FSTD  VSTR

#define VNEGvfp(_Dd,_Dm) do {                                           \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int dflag = DFlag(_Dd); NanoAssert(dflag==DFlag(_Dm));          \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                 \
		int DmEnc; encodeRegM(_Dm,true , true , false);                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xEB10A40) | DdEnc | DmEnc | dflag ); \
		asm_output("fneg%c %s,%s",dflag?'d':'s', gpn(_Dd), gpn(_Dm));   \
	 } while (0)

#define VNEGneon(_Qd,_Qm) do {                                          \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                 \
		int DmEnc; encodeRegM(_Qm,false, false, true);                                 \
		*(--_nIns) = (NIns)( 0xF3B907C0 | DdEnc | DmEnc );              \
		asm_output("fnegq %s,%s", gpn(_Qd), gpn(_Qm));                  \
	 } while (0)


#define GenericVfp(NAME,_Dd,_Dn,_Dm,SPECIFIC_ENC) do {                  \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int dflag = DFlag(_Dd); NanoAssert(dflag==DFlag(_Dm));NanoAssert(dflag==DFlag(_Dn)); \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                 \
		int DmEnc; encodeRegM(_Dm,true , true , false);                                 \
		int DnEnc; encodeRegN(_Dn,true , true , false);                                 \
		*(--_nIns) = (NIns)( COND_AL | SPECIFIC_ENC | DnEnc | DdEnc | DmEnc | dflag);        \
		asm_output(#NAME"%c %s,%s,%s",dflag?'d':'s', gpn(_Dd), gpn(_Dn), gpn(_Dm));          \
	 } while (0)

#define GenericNeon(NAME,_Qd,_Qn,_Qm,SPECIFIC_ENC) do {                 \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                 \
		int DmEnc; encodeRegM(_Qm,false, false, true);                                 \
		int DnEnc; encodeRegN(_Qn,false, false, true);                                 \
		*(--_nIns) = (NIns)( (0xF2U << 24) | (SPECIFIC_ENC) | DdEnc | DnEnc| DmEnc );         \
		asm_output(#NAME" %s,%s,%s", gpn(_Qd), gpn(_Qn), gpn(_Qm));     \
	} while (0)


#define VADDvfp(_Dd,_Dn,_Dm)  GenericVfp(fadd,_Dd,_Dn,_Dm,0xE300A00)
#define VSUBvfp(_Dd,_Dn,_Dm)  GenericVfp(fsub,_Dd,_Dn,_Dm,0xE300A40)
#define VMULvfp(_Dd,_Dn,_Dm)  GenericVfp(fmul,_Dd,_Dn,_Dm,0xE200A00)
#define VDIVvfp(_Dd,_Dn,_Dm)  GenericVfp(fdiv,_Dd,_Dn,_Dm,0xE800A00)

#define VADDneon(_Qd,_Qn,_Qm) GenericNeon(vadd,_Qd,_Qn,_Qm, 0xD40)
#define VSUBneon(_Qd,_Qn,_Qm) GenericNeon(vsub,_Qd,_Qn,_Qm, 0x200D40)
#define VMULneon(_Qd,_Qn,_Qm) GenericNeon(vmul,_Qd,_Qn,_Qm, 0x1000D50)
  /* neon doesn't have VDIV; we'll do it on the components */
#define VDIVneon(_Qd,_Qn,_Qm) do{                                       \
			NanoAssert(IsFpQReg(_Qn) && IsFpQReg(_Qd) && IsFpQReg(_Qm));\
			Register Sn = S0 + (FpQRegNum(_Qn) << 2);                   \
			Register Sm = S0 + (FpQRegNum(_Qm) << 2);                   \
			Register Sd = S0 + (FpQRegNum(_Qd) << 2);                   \
			VDIVvfp(Sd,Sn,Sm); Sn = Sn + 1; Sm = Sm + 1; Sd = Sd + 1;   \
			VDIVvfp(Sd,Sn,Sm); Sn = Sn + 1; Sm = Sm + 1; Sd = Sd + 1;   \
			VDIVvfp(Sd,Sn,Sm); Sn = Sn + 1; Sm = Sm + 1; Sd = Sd + 1;   \
			VDIVvfp(Sd,Sn,Sm);                                          \
		} while (0) 

#define VADD(_Dd,_Dn,_Dm) if(IsFpQReg(_Dd)) VADDneon(_Dd,_Dn,_Dm); else VADDvfp(_Dd,_Dn,_Dm);
#define VSUB(_Dd,_Dn,_Dm) if(IsFpQReg(_Dd)) VSUBneon(_Dd,_Dn,_Dm); else VSUBvfp(_Dd,_Dn,_Dm);
#define VMUL(_Dd,_Dn,_Dm) if(IsFpQReg(_Dd)) VMULneon(_Dd,_Dn,_Dm); else VMULvfp(_Dd,_Dn,_Dm);
#define VDIV(_Dd,_Dn,_Dm) if(IsFpQReg(_Dd)) VDIVneon(_Dd,_Dn,_Dm); else VDIVvfp(_Dd,_Dn,_Dm);
#define VNEG(_Dd, _Dm)    if(IsFpQReg(_Dd)) VNEGneon(_Dd,_Dm);     else VNEGvfp(_Dd,_Dm);
#define VCEQ(_Qd,_Qn,_Qm) GenericNeon(vceq.f32,_Qd,_Qn,_Qm,0xE40)

#define VMOVN(_Dd,_Qm)  do { \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int DdEnc; encodeRegD(_Dd,false, true , false);                                 \
		int DmEnc; encodeRegM(_Qm,false, false, true);                                 \
		*(--_nIns) = (NIns)( 0xF3B60200 | DdEnc | DmEnc );              \
		asm_output("vmovn.i32 %s,%s", gpn(_Dd), gpn(_Qm));              \
	} while (0)

#define VMVN(_Dd,_Dm)  do { \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int DdEnc; encodeRegD(_Dd,false, true , false);                                 \
		int DmEnc; encodeRegM(_Dm,false, true , false);                                 \
		*(--_nIns) = (NIns)( 0xF3B00580 | DdEnc | DmEnc );              \
		asm_output("vmvn %s,%s", gpn(_Dd), gpn(_Dm));                   \
	} while (0)

#define VPMINU16(_Dd,_Dn,_Dm)  do { \
		underrunProtect(4);                                             \
		NanoAssert(ARM_VFP);                                            \
		int DdEnc; encodeRegD(_Dd,false, true , false);                                 \
		int DmEnc; encodeRegM(_Dm,false, true , false);                                 \
		int DnEnc; encodeRegN(_Dn,false, true , false);                                 \
		*(--_nIns) = (NIns)( 0xF3100A10 | DnEnc | DdEnc | DmEnc );      \
		asm_output("vpmin.u16 %s,%s,%s", gpn(_Dd), gpn(_Dn), gpn(_Dm)); \
	} while (0)
	
#define VDUP(_Qd,_Sm) do {\
		underrunProtect(4);                                                   \
		NanoAssert(ARM_VFP);                                                  \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                       \
		int M,Vm; NanoAssert(IsFpSReg(_Sm));                                  \
		split_sreg_enc(_Sm,M,Vm);                                             \
   /* M is part of th 'imm4' encoding; the first bit. the rest is 0x4 (100)*/ \
   /* The actual M is always 0 - Sn registers never get to D16-31 range    */ \
		*(--_nIns) = (NIns)( 0xF3B40C40 | DdEnc | Vm | (M<<19) );             \
		asm_output("vdup.32 %s,%s", gpn(_Qd), gpn(_Sm));                      \
	} while (0)


#define FMSTAT() do {                                                       \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		*(--_nIns) = (NIns)( COND_AL | 0x0EF1FA10);                         \
		asm_output("fmstat");                                               \
	} while (0)

#define VCMP(_Dd,_Dm,_E) do {                                               \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		int dflag = DFlag(_Dd); NanoAssert(dflag==DFlag(_Dm));              \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                     \
		int DmEnc; encodeRegM(_Dm,true , true , false);                                     \
		NanoAssert(((_E)==0) || ((_E)==1));                                 \
		*(--_nIns) = (NIns)( COND_AL | (0xEB4<<16) | DdEnc | 0xA40 | ((_E)<<7) | dflag | DmEnc ); \
		asm_output("fcmp%s%c %s,%s", (((_E)==1)?"e":""),dflag?'d':'s', gpn(_Dd), gpn(_Dm));       \
	} while (0)

#define VCMPZ(_Dd) do {                                                     \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		int DdEnc; encodeRegD(_Dd,false, true , false);                                     \
		*(--_nIns) = (NIns)( COND_AL | (0xEB5<<16) | DdEnc | 0xBC0 );       \
		asm_output("vcmpz %s", gpn(_Dd) );                                  \
	} while (0)

#define VMOV_cond(_cond,_Dd,_Dm) do {                                       \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		NanoAssert(IsCond(_cond));                                          \
		int dflag = DFlag(_Dd); NanoAssert(dflag==DFlag(_Dm));              \
		int DdEnc; encodeRegD(_Dd,true , true , false);                                     \
		int DmEnc; encodeRegM(_Dm,true , true , false);                                     \
		*(--_nIns) = (NIns)( (((unsigned int)_cond)<<28) | (0xEB0<<16) | DdEnc | 0xA40 | DmEnc | dflag );       \
		asm_output("fcpy%c%s %s,%s", dflag?'d':'s', condNames[_cond], gpn(_Dd), gpn(_Dm));        \
	 } while (0)

#define FCPYS(_Sd,_Sm)      VMOV_cond(AL,_Sd,_Sm)
#define FCPYD(_Dd,_Dm)      VMOV_cond(AL,_Dd,_Dm)
#define FCPYD_cond          VMOV_cond
#define FCPYQ(_Qd,_Qm)  do {                                                \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP); /*NEON!*/                                      \
		int DdEnc; encodeRegD(_Qd,false, false, true);                                     \
		int DmEnc; encodeRegM(_Qm,false, false, true);                                     \
		int DnEnc; encodeRegN(_Qm,false, false, true);                                     \
		*(--_nIns) = (NIns)( (0xF2200150) | DdEnc  | DmEnc  | DnEnc );      \
		asm_output("fcpyq %s,%s", gpn(_Qd), gpn(_Qm));                      \
	} while (0)

#define FMSR(_Sn,_Rd) do {                                                  \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		NanoAssert( IsGpReg(_Rd) );                                         \
		int DnEnc = 0; encodeRegN(_Sn,true , false, false);                                     \
		*(--_nIns) = (NIns)( COND_AL | (0xE0<<20) |((REGNUM(_Rd))<<12) | 0xA10 | DnEnc ); \
		asm_output("fmsr %s,%s", gpn(_Sn), gpn(_Rd));                       \
	} while (0)

#define FMRS(_Rd,_Sn) do {                                                  \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		NanoAssert( IsGpReg(_Rd) );                                         \
		int DnEnc; encodeRegN(_Sn,true , false, false);                                     \
		*(--_nIns) = (NIns)( COND_AL | (0xE1<<20) | ((REGNUM(_Rd))<<12) | 0xA10 | DnEnc ); \
		asm_output("fmrs %s,%s", gpn(_Rd), gpn(_Sn));                       \
	} while (0)


/* Conversions: to/from int(signed, unsigned), to/from single/double precision */

#define VCVTx(NAME,_Rd,_Rm,SPECIFIC_ENC,check_dst) do {                     \
		underrunProtect(4);                                                 \
		NanoAssert(ARM_VFP);                                                \
		int dflag = DFlag((check_dst?_Rd:_Rm));                             \
		int DdEnc; encodeRegD(_Rd,true , true , false);                                     \
		int DmEnc; encodeRegM(_Rm,true , true , false);                                     \
		*(--_nIns) = (NIns)( COND_AL | (0xEB<<20) | SPECIFIC_ENC | dflag | DdEnc | DmEnc ); \
		asm_output(#NAME" %s,%s",gpn(_Rd), gpn(_Rm));                       \
	} while(0)
#define VCVT(A,B,C,D) VCVTx(A,B,C,D,true)

#define FCVTDS(_Dd,_Sm) do {                                               \
		NanoAssert(IsFpDReg(_Dd) && IsFpSReg(_Sm));                        \
		VCVTx(fcvtds,_Dd,_Sm,0x70AC0,false);                               \
	} while (0)

#define FCVTSD(_Sd,_Dm) do {                                               \
		NanoAssert(IsFpSReg(_Sd) && IsFpDReg(_Dm));                        \
		VCVTx(fcvtsd,_Sd,_Dm,0x70AC0,false);                               \
	} while (0)

#define FSITOD(_Dd,_Sm) do {                                               \
		NanoAssert(IsFpDReg(_Dd) && IsFpSReg(_Sm));                        \
		VCVT(fsitod,_Dd,_Sm,0x80AC0);                                      \
	} while (0)

#define FUITOD(_Dd,_Sm) do {                                               \
		NanoAssert(IsFpDReg(_Dd) && IsFpSReg(_Sm));                        \
		VCVT(fuitod,_Dd,_Sm,0x80A40);                                      \
	} while (0)

#define FSITOS(_Sd,_Sm) do {                                              \
		NanoAssert(IsFpSReg(_Sd) && IsFpSReg(_Sm));                       \
		VCVT(fsitos,_Sd,_Sm,0x80AC0);                                     \
	} while (0)

//  unsigned int, to single-precision float
#define FUITOS(_Sd,_Sm) do {                                              \
		NanoAssert(IsFpSReg(_Sd) && IsFpSReg(_Sm));                       \
		VCVT(fuitos,_Sd,_Sm,0x80A40);                                     \
	} while (0)

//  Double to signed int, stored in Sn reg
#define FTOSID(_Sd,_Dm)do {                                               \
		NanoAssert(IsFpSReg(_Sd) && IsFpDReg(_Dm));                       \
		VCVTx(ftosid,_Sd,_Dm,0xD0A40,false);                              \
	} while (0)

//  Single-precision float to signed int, stored in Sn reg
#define FTOSIS(_Sd,_Sm)do {                                               \
		NanoAssert(IsFpSReg(_Sd) && IsFpSReg(_Sm));                       \
		VCVTx(ftosis,_Sd,_Sm,0xD0A40,false);                              \
	} while (0)

} // namespace nanojit

#endif // __nanojit_NativeARM_Macros_
