/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nanojit.h"

#include "NativePPC.h"

#if defined FEATURE_NANOJIT && defined NANOJIT_PPC

#ifdef NANOJIT_64BIT
#error "Only 32 bit mode supported"
#endif

namespace nanojit
{

	const char *regNames[] = {
		"r0",  "sp",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
		"r8",  "r9",  "r10", "r11", "r12", "r13", "r14", "r15",
		"r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
		"r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31",
		"f0",  "f1",  "f2",  "f3",  "f4",  "f5",  "f6",  "f7",
		"f8",  "f9",  "f10", "f11", "f12", "f13", "f14", "f15",
		"f16", "f17", "f18", "f19", "f20", "f21", "f22", "f23",
		"f24", "f25", "f26", "f27", "f28", "f29", "f30", "f31"
	};

	const char *bitNames[] = { "lt", "gt", "eq", "so" };

	#define TODO(x) do{ avmplus::AvmLog(#x##"\n"); NanoAssertMsgf(false, "%s", #x); } while(0)
	#define INFO(x) do{ avmplus::AvmLog(#x##"\n"); } while (0)

#if defined(NANOJIT_PPC_ENABLE_SWAP)
	// This rather overgrown macro does multiple things:
	// - Emits an instruction
	// - Tracks opcode history for our optimizer
	// - Actually *is* our optimizer: if the previous instruction is
	//   swappable and this instruction has no register dependencies on
	//   it, swap it up with the previous instruction so we can get more
	//   sillycon on board. Also make sure the previous instruction is what
	//   we think it is: we could have been "overrun" into another page.
	//   This is less effective for aggressively reordered CPUs like the 970
	//   and POWER7, but is helpful for 750s and 74xx, and also in-order
	//   implementations like POWER6.
	// - Special case for stores: turn st cmp bc into cmp st bc so we can get
	//   the cmp for "free" as the store is independent.
	// - Special case for overflow instructions: turn st (addo,mullwo,subfo)
	//   mcrxr bc into (a,m,s) mcrxr st bc for the same reason.
	// We abuse the macro for this purpose to make optimization as
	// aggressive and ubiquitous as possible across individual LIR calls.
	// Fortunately, this is trivial for a decent compiler to optimize since
	// there are lots of constants and thus often clearly dead code for
	// certain instructions.
	void Assembler::doSwap(PpcOpcode op, Register r1, Register r2, Register r3, int32_t ins) {
		NIns swapme = (NIns)0;
		bool doubleswap = false;
		if (!_doNotSwap && _lastOpcode.reg1 != NoSwap && r1 != NoSwap &&
			_lastOpcode.instr == *(_nIns)) {
				if (op == PPC_stb || op == PPC_stbx ||
					op == PPC_sth || op == PPC_sthx || op == PPC_sthbrx ||
					op == PPC_stw || op == PPC_stwx || op == PPC_stwu || op == PPC_stwux || op == PPC_stwbrx ||
					op == PPC_std || op == PPC_stdu || op == PPC_stdux || op == PPC_stdx ||
					op == PPC_stfd || op == PPC_stfdx || op == PPC_stfs || op == PPC_stfsx) {
						if (_lastOpcode.opcode == PPC_cmp ||
							_lastOpcode.opcode == PPC_cmpi ||
							_lastOpcode.opcode == PPC_cmpl ||
							_lastOpcode.opcode == PPC_cmpli) {
								_lastOpcode.reg1 = NoReg;
								_lastOpcode.reg2 = NoReg;
								_lastOpcode.reg3 = NoReg;
						}
						if ((_lastOpcode.opcode == PPC_addo ||
							_lastOpcode.opcode == PPC_mullwo ||
							_lastOpcode.opcode == PPC_subfo) &&
							_lastOpcode.reg1 != r1 &&
							((*(_nIns+1) & 0xfc7fffff) == PPC_mcrxr)) {
								asm_output("; double swap against overflow + mcrxr");
								doubleswap = true;
								*(_nIns-1) = *(_nIns);
								*(_nIns) = *(_nIns+1);
								*(_nIns+1) = ins;
								_nIns--;
						}
				}
				if (!doubleswap && (_lastOpcode.reg1 == NoReg ||
					(_lastOpcode.reg1 != r1 &&
					_lastOpcode.reg1 != r2 &&
					_lastOpcode.reg1 != r3)) &&
					(_lastOpcode.reg2 == NoReg ||
					(_lastOpcode.reg2 != r1 &&
					_lastOpcode.reg2 != r2 &&
					_lastOpcode.reg2 != r3)) &&
					(_lastOpcode.reg3 == NoReg ||
					(_lastOpcode.reg3 != r1 &&
					_lastOpcode.reg3 != r2 &&
					_lastOpcode.reg3 != r3))) {
						asm_output("; swapped @ %p\n", _nIns);
						swapme = *(_nIns++);
				}
		}

		if (!doubleswap)
			*(--_nIns) = (NIns)(ins);

		if(swapme) {
			*(--_nIns) = swapme;
			_lastOpcode.reg1 = NoSwap;
		} else {
			_lastOpcode.opcode = (NIns)op;
			_lastOpcode.reg1 = (doubleswap) ? NoSwap : r1;
			_lastOpcode.reg2 = r2;
			_lastOpcode.reg3 = r3;
			_lastOpcode.instr = ins;
		}
	}
#endif


	/*
	 * sp+0     saved sp
	 * sp+4     saved lr
	 */

	NIns* Assembler::genEntryPoint(RegisterMask saveMask) {

		NanoAssert(saveMask == 0);
		NanoUnusedVariable(saveMask);

		STW(R0, 12, SP);
		STWU(SP, -8, SP);
		MFLR(R0);
		return _nIns;
	}

	NIns* Assembler::genExitPoint(RegisterMask restoreMask) {

		NanoAssert(restoreMask == 0);
		NanoUnusedVariable(restoreMask);

		BLR();
		ADDI(SP, SP, 8);
		MTLR(R0);
		LWZ(R0, 12, SP);
		return _nIns;
	}

	void Assembler::genPrologue() {
	}

	void Assembler::asm_load32(LIns *ins) {
		LIns* base = ins->oprnd1();
		int d = ins->disp();
		Register rr = prepareResultReg(ins, GpRegs);
		Register ra = getBaseReg(base, d, GpRegs);

		LOpcode op = ins->opcode();

		switch(op) {
			case LIR_ldc2i:
			case LIR_lduc2ui:
				if (op == LIR_ldc2i)
					EXTSB(rr, rr);
				if (isS16(d)) {
					LBZ(rr, d, ra);
				} else {
					LBZX(rr, ra, R0); // rr = [ra+R0]
					asm_li(R0,d);
				}
				break;

			case LIR_lds2i:
				// these are expected to be 2 or 4-bytealigned
				if (isS16(d)) {
					LHA(rr, d, ra);
				} else {
					LHAX(rr, ra, R0); // rr = [ra+R0]
					asm_li(R0, d);
				}
				break;

			case LIR_ldus2ui:
				// these are expected to be 2 or 4-byte aligned
				if (isS16(d)) {
					LHZ(rr, d, ra);
				} else {
					LHZX(rr, ra, R0); // rr = [ra+R0]
					asm_li(R0, d);
				}
				break;

			case LIR_ldi:
				// these are expected to be 4-byte aligned
				if (isS16(d)) {
					LWZ(rr, d, ra);
				} else {
					LWZX(rr, ra, R0); // rr = [ra+R0]
					asm_li(R0,d);
				}
				break;

			case LIR_ldri:
				if (d == 0) {
					LWBRX(rr, R0, ra);
				} else {
					LWBRX(rr, ra, R0);
					asm_li(R0, d);
				}
				break;

			case LIR_ldrs2i:
			case LIR_ldrus2ui:
				if (op == LIR_ldrs2i)
					EXTSH(rr, rr);

				if (d == 0) {
					LHBRX(rr, R0, ra);
				} else {
					LHBRX(rr, ra, R0);
					asm_li(R0, d);
				}
				break;

			default:
				NanoAssertMsg(0, "asm_load32 should never receive this LIR opcode");
				return;
		}

		freeResourcesOf(ins);
	}

	void Assembler::asm_store32(LOpcode op, LIns *value, int32_t dr, LIns *base) {

		// XXX?
		switch (op) {
			case LIR_sti:
			case LIR_stri:
			case LIR_sti2c:
			case LIR_sti2s:
			case LIR_stri2s:
				// handled by mainline code below for now
				break;
			default:
				NanoAssertMsg(0, "asm_store32 should never receive this LIR opcode");
				return;
		}

		Register rs = findRegFor(value, GpRegs);
		Register ra = value == base ? rs : getBaseReg(base, dr, GpRegs & ~rmask(rs));

		switch (op) {
			case LIR_sti:
				if (isS16(dr)) {
					STW(rs, dr, ra);
				} else {
					STWX(rs, ra, R0);
					asm_li(R0, dr);
				}
				break;

			case LIR_sti2c:
				if (isS16(dr)) {
					STB(rs, dr, ra);
				} else {
					STBX(rs, ra, R0);
					asm_li(R0, dr);
				}
				break;

			case LIR_sti2s:
				if (isS16(dr)) {
					STB(rs, dr, ra);
				} else {
					STHX(rs, ra, R0);
					asm_li(R0, dr);
				}
				break;

			case LIR_stri:
				if (dr == 0) {
					STWBRX(rs, R0, ra);
				} else {
					STWBRX(rs, ra, R0);
					asm_li(R0, dr);
				}
				break;

			case LIR_stri2s:
				if (dr == 0) {
					STHBRX(rs, R0, ra);
				} else {
					STHBRX(rs, ra, R0);
					asm_li(R0, dr);
				}
				break;
		}

	}

	void Assembler::asm_load64(LIns *ins) {

		// XXX?
		switch (ins->opcode()) {
			case LIR_ldf2d:
			case LIR_ldd:
			CASE64(LIR_ldq:)
				// handled by mainline code below for now
				break;
			default:
				NanoAssertMsg(0, "asm_load64 should never receive this LIR opcode");
				return;
		}

		LIns* base = ins->oprnd1();
		Register rr = prepareResultReg(ins, FpRegs);

		int dr = ins->disp();
		Register ra = getBaseReg(base, dr, GpRegs);

		// FPR
	#if !PEDANTIC
		if (isS16(dr)) {
			if (ins->opcode() == LIR_ldf2d)
				LFS(rr, dr, ra);
			else
				LFD(rr, dr, ra);
			return;
		}
	#endif

		// general case FPR load
		if (ins->opcode() == LIR_ldf2d)
			LFSX(rr, ra, R0);
		else
			LFDX(rr, ra, R0);
		asm_li(R0, dr);

		freeResourcesOf(ins);
	}

	void Assembler::asm_li(Register r, int32_t imm) {
		if (isS16(imm)) {
			LI(r, imm);
			return;
		}
		if ((imm & 0xffff) == 0) {
			imm = uint32_t(imm) >> 16;
			LIS(r, imm);
			return;
		}
		asm_li32(r, imm);
	}

	void Assembler::asm_li32(Register r, int32_t imm) {
		// general case. One day, a real three-way adder will come and
		// wash all the scum off the streets.
		ORI(r,r,imm);
		LIS(r, imm>>16);
	}

	void Assembler::asm_li64(Register r, uint64_t imm) {
		NanoAssert(0); // cannot be called in 32 bit mode
	}

	void Assembler::asm_store64(LOpcode op, LIns *value, int32_t dr, LIns *base) {
		NanoAssert(value->isQorD());

		// XXX?
		switch (op) {
			case LIR_std:
			CASE64(LIR_stq:)
			case LIR_std2f:
				// handled by mainline code below for now
				break;
			default:
				NanoAssertMsg(0, "asm_store64 should never receive this LIR opcode");
				return;
		}

		Register ra = getBaseReg(base, dr, GpRegs);

		// general case for any value
		// on 32bit cpu's, we only use store64 for doubles
		Register rs = findRegFor(value, FpRegs);

	#if !PEDANTIC
		if (isS16(dr)) {
			// short offset
			if (op == LIR_std2f) {
				STFS(rs, dr, ra);
			} else {
				STFD(rs, dr, ra);
			}
			return;
		}
	#endif

		// general case for any offset
		if (op == LIR_std2f) {
			STFSX(rs, ra, R0);
		} else {
			STFDX(rs, ra, R0);
		}
		asm_li(R0, dr);
	}

	void Assembler::asm_cond(LIns *ins) {
		LOpcode op = ins->opcode();
		LIns *a = ins->oprnd1();
		LIns *b = ins->oprnd2();
		SwapDisable();
		ConditionRegister cr = CR0;
		Register r = prepareResultReg(ins, GpRegs);
		switch (op) {
		case LIR_eqi: case LIR_eqd:
		CASE64(LIR_eqq:)
			EXTRWI(r, r, 1, 4*cr+COND_eq); // extract CR0.eq
			MFCR(r);
			break;
		case LIR_lti: case LIR_ltui:
		case LIR_ltd: case LIR_led:
		CASE64(LIR_ltq:) CASE64(LIR_ltuq:)
			EXTRWI(r, r, 1, 4*cr+COND_lt); // extract CR0.lt
			MFCR(r);
			break;
		case LIR_gti: case LIR_gtui:
		case LIR_gtd: case LIR_ged:
		CASE64(LIR_gtq:) CASE64(LIR_gtuq:)
			EXTRWI(r, r, 1, 4*cr+COND_gt); // extract CR0.gt
			MFCR(r);
			break;
		case LIR_lei: case LIR_leui:
		CASE64(LIR_leq:) CASE64(LIR_leuq:)
			EXTRWI(r, r, 1, 4*cr+COND_eq); // extract CR0.eq
			MFCR(r);
			CROR(CR0, eq, lt, eq);
			break;
		case LIR_gei: case LIR_geui:
		CASE64(LIR_geq:) CASE64(LIR_geuq:)
			EXTRWI(r, r, 1, 4*cr+COND_eq); // select CR0.eq
			MFCR(r);
			CROR(CR0, eq, gt, eq);
			break;
		default:
			verbose_only(outputf("%s", getLirName(ins->opcode()));)
			TODO(asm_cond);
			break;
		}
		SwapEnable();
		asm_cmp(op, a, b, cr);

		freeResourcesOf(ins);
	}

	void Assembler::asm_condd(LIns *ins) {
		asm_cond(ins);
	}

	// cause sign extension to test bits.  ptrdiff_t is a signed,
	// pointer-sized int
	static inline bool isS14(ptrdiff_t d) {
		const int shift = sizeof(ptrdiff_t) * 8 - 14; // 18 or 50
		return ((d << shift) >> shift) == d;
	}

	NIns* Assembler::asm_branch(const NIns * targ) {
		SwapDisable();
		underrunProtect(8);

		// powerpc offsets are based on the address of the branch instruction
		ptrdiff_t bd = targ - (_nIns-1);

		if (targ == NULL) {
			B(0);
		} else if (isS24(bd)) {
			B(bd);
		} else {
			TRAP();
			NanoAssert(0);
		}

		SwapEnable();

		return _nIns;
	}

	void Assembler::asm_branch_cond(bool onfalse, LIns* cond, const NIns* targ, bool patchable, Branches* branches) {
		SwapDisable();
		underrunProtect(16);

		NanoUnusedVariable(patchable);

		LOpcode condop = cond->opcode();
		NanoAssert(cond->isCmp());

		ptrdiff_t bd = targ - (_nIns-1); // *instructions* (3 nops + bc = 4)
		const uint32_t likely = 0;

		if ((targ == NULL) || !isS14(bd)) {
			// can't fit conditional branch offset into 14 bits, but
			// we can fit in 24, so invert the condition and branch
			// around an unconditional jump
			verbose_only(verbose_outputf("%p:", _nIns);)
			NIns *skip = _nIns;
			B(bd);
			branches->addBranch(_nIns, targ);

			targ = skip;
			onfalse = !onfalse;
			bd = skip - (_nIns-1); // this is correct.
			NanoAssert(isS14(bd));
		}

		ConditionRegister cr = CR0;
		switch (cond->opcode()) {
		case LIR_eqi:
		case LIR_eqd:
		CASE64(LIR_eqq:)
			if (onfalse) BNE(cr,bd,likely); else BEQ(cr,bd,likely);
			break;
		case LIR_nei:
		case LIR_ned:
		CASE64(LIR_neq:)
			if (onfalse) BEQ(cr,bd,likely); else BNE(cr,bd,likely);
			break;
		case LIR_lti: case LIR_ltui:
		case LIR_ltd: case LIR_led:
		CASE64(LIR_ltq:) CASE64(LIR_ltuq:)
			if (onfalse) BNL(cr,bd,likely); else BLT(cr,bd,likely);
			break;
		case LIR_lei: case LIR_leui:
		CASE64(LIR_leq:) CASE64(LIR_leuq:)
			if (onfalse) BGT(cr,bd,likely); else BLE(cr,bd,likely);
			break;
		case LIR_gti: case LIR_gtui:
		case LIR_gtd: case LIR_ged:
		CASE64(LIR_gtq:) CASE64(LIR_gtuq:)
			if (onfalse) BNG(cr,bd,likely); else BGT(cr,bd,likely);
			break;
		case LIR_gei: case LIR_geui:
		CASE64(LIR_geq:) CASE64(LIR_geuq:)
			if (onfalse) BLT(cr,bd,likely); else BGE(cr,bd,likely);
			break;
		default:
			verbose_only(outputf("%s", getLirName(cond->opcode()));)
			TODO(unknown_cond);
			break;
		}

		branches->addBranch(_nIns, targ);

		SwapEnable();

		asm_cmp(condop, cond->oprnd1(), cond->oprnd2(), CR0);
	}

	void Assembler::asm_cmp(LOpcode condop, LIns *a, LIns *b, ConditionRegister cr) {
		RegisterMask allow = isCmpDOpcode(condop) ? FpRegs : GpRegs;
		Register ra = findRegFor(a, allow);

	#if !PEDANTIC
		if (b->isImmI() && b->canOptimize()) {
			int32_t d = b->immI();
			if (isS16(d)) {
				if (isCmpSIOpcode(condop)) {
					CMPWI(cr, ra, d);
					return;
				}
			}
			if (isU16(d)) {
				if (isCmpUIOpcode(condop)) {
					CMPLWI(cr, ra, d);
					return;
				}
			}
		}
	#endif

		// general case
		Register rb = b==a ? ra : findRegFor(b, allow & ~rmask(ra));
		if (isCmpSIOpcode(condop)) {
			CMPW(cr, ra, rb);
		}
		else if (isCmpUIOpcode(condop)) {
			CMPLW(cr, ra, rb);
		}
		else if (isCmpDOpcode(condop)) {
			// set the lt/gt bit for fle/fge.  We don't do this for
			// int/uint because in those cases we can invert the branch condition.
			// for float, we can't because of unordered comparisons
			if (condop == LIR_led)
				CROR(cr, lt, lt, eq); // lt = lt|eq
			else if (condop == LIR_ged)
				CROR(cr, gt, gt, eq); // gt = gt|eq
			FCMPU(cr, ra, rb);
		}
		else {
			TODO(asm_cmp);
		}

	}

	void Assembler::asm_gp_copy(Register d, Register s) {
		MR(d, s);
	}

	void Assembler::asm_nongp_copy(Register d, Register s) {
		// PPC doesn't support any GPR<->FPR moves
		NanoAssert((rmask(d) & FpRegs) && (rmask(s) & FpRegs));
		FMR(d, s);
	}

	void Assembler::asm_restore(LIns *i, Register r) {
		int d;
		if (i->isImmI() && i->canOptimize()) {
			asm_li(r, i->immI());
		}
		else {
			NanoAssert(!RegAlloc::canRemat(i));
			d = findMemFor(i);
			if (i->isI()) {
				switch (i->getArSize()) {
					case 4:		LWZ(r, d, FP); break;
					case 2:		LHZ(r, d, FP); break;
					case 1:		LBZ(r, d, FP); break;
					default:	NanoAssert(0); break;
				}
			} else if (IsFpReg(r)) {
				NanoAssert(i->isQorD());
				LFD(r, d, FP);
			} else if (i->isQorD()) {
				NanoAssert(IsGpReg(r));
				LD(r, d, FP);
			}
		}
	}

	void Assembler::asm_immi(LIns *ins) {
		Register rr = prepareResultReg(ins, GpRegs);
		if (ins->getGenFlags() &&  (ins->flagRequested() != 0))
		{
			NanoAssert((ins->flagRequested() & (ArithFlag_Overflow | ArithFlag_Carry | ArithFlag_Borrow)) == 0);
			asm_force_gen_flags(ins);
		}

		asm_li(rr, ins->immI());
		freeResourcesOf(ins);
	}

	void Assembler::asm_neg_abs(LIns *ins) {
		Register rr = prepareResultReg(ins, FpRegs);
		LIns* lhs = ins->oprnd1();
		// We can clobber the result register for "free."
		Register ra = ( !lhs->isInReg()
					  ? findSpecificRegFor(lhs, rr)
					  : findRegFor(lhs, FpRegs) );

		FNEG(rr,ra);
	}

	void Assembler::asm_call(LIns* ins, const NIns *address) {

		SwapDisable();

		ParameterValue * callParams = ins->callArgs();
		unsigned int paramCount = ins->callArgc();
		RegisterMask evictMask = ins->callVolatileMask();

		switch (ins->opcode())
		{
		case LIR_calli:
			CASE64(LIR_callq:)
			{
				Register r = prepareResultReg(ins, rmask(R3));
				evictMask &= ~rmask(r);
				break;
			}

		case LIR_callf:
		case LIR_calld:
			{
				Register r = prepareResultReg(ins, rmask(F0));
				evictMask &= ~rmask(r);
				break;
			}

		case LIR_callv:
			break;

		default:
			NanoAssert(0);
		}

		evictSomeActiveRegs(evictMask);

		NanoAssert(address != 0);

		
		br(address, 1);
		

		NanoAssert(address != 0);

		if (ins->opcode() != LIR_callv)
		{
			freeResourcesOf(ins);
		}

		reconcileParamRegs(callParams, paramCount);

		SwapEnable();

	}

	void Assembler::asm_spill(const LIns* insToSpill) {

		int d = arDisp(insToSpill->getArIndex());
		NanoAssert(d >= 0);
		NanoAssert(insToSpill->isInReg());

		Register rr = insToSpill->getReg();

		if (IsGpReg(rr)) {
			NanoAssert(insToSpill->isI());
			switch (insToSpill->getArSize()) {
				case 4:		STW(rr, d, FP); break;
				case 2:		STH(rr, d, FP); break;
				case 1:		STB(rr, d, FP); break;
				default:	NanoAssert(0); break;
			}
		} else {
			NanoAssert(IsFpReg(rr));
			STFD(rr, d, FP);
		}
	}

	bool Assembler::asm_arith_imm(LIns* ins) {
		LIns* lhs = ins->oprnd1();
		LIns* rhs = ins->oprnd2();
		int32_t rhsc = rhs->immI();

		// register assignment was already done in asm_arith
		Register rr = prepareResultReg(ins, GpRegs);
		Register ra = findRegFor(lhs, GpRegs & ~rmask(rr));
		LOpcode op = ins->opcode();

		if (isS16(rhsc)) {
			// ppc arith immediate ops sign-extend the imm16    value
			switch (op) {
			case LIR_addi:
			CASE64(LIR_addq:)
				ADDI(rr, ra, rhsc);
				return true;
			case LIR_subi:
				ADDI(rr, ra, -rhsc);
				return true;
			case LIR_muli:
				// MULLI can take up to 5 cycles in the worst case.
				// Let's try desperately to avoid it with some
				// strength-reduced chains of addition. We might even
				// write up a shifter at some point, but signage is
				// a little complex.
				switch(rhsc) {
				case 0:
				case 1:
					// uh ...
					INFO(multiply_by_zero_shouldnt_this_be_optimized_out);
					break; // fall through
				case 2:
					// Add itself to itself.
					ADD(rr, ra, ra);
					return true;
				case 3:
					// This is on SunSpider, btw.
					// Save a copy in R0 if rr == ra because it will be
					// clobbered.
					if (rr == ra) {
						ADD(rr, rr, R0);
					} else {
						ADD(rr, rr, ra);
					}
					// Add itself to itself.
					ADD(rr, ra, ra);
					if (rr == ra) {
						MR(R0, ra);
					}
					return true;
				case 4:
					// Add itself to itself and itself to itself.
					// XXX A signed shift left might be faster.
					ADD(rr, rr, rr);
					ADD(rr, ra, ra);
					return true;
				case 5:
					// Save a copy in R0 if rr == ra because it will be
					// clobbered.
					if (rr == ra) {
						ADD(rr, rr, R0);
					} else {
						ADD(rr, rr, ra);
					}
					// Add itself to itself and itself to itself.
					ADD(rr, rr, rr);
					ADD(rr, ra, ra);
					if (rr == ra) {
						MR(R0, ra);
					}
					return true;
				case 8:
					// Add itself to itself and itself to itself, then
					// itself to itself.
					// XXX A signed shift left likely will be faster.
					ADD(rr, rr, rr);
					ADD(rr, rr, rr);
					ADD(rr, ra, ra);
					return true;
				case 10:
					// Like 5, but then a post add. Worst case is
					// still 5 cycles, but it might be taken out of order
					// and we can probably beat that then.
					ADD(rr, rr, rr);
					if (rr == ra) {
						ADD(rr, rr, R0);
					} else {
						ADD(rr, rr, ra);
					}
					// Add itself to itself and itself to itself.
					ADD(rr, rr, rr);
					ADD(rr, ra, ra);
					if (rr == ra) {
						MR(R0, ra);
					}
					return true;
				case 16:
					// Add itself to itself and itself to itself, then
					// itself to itself, and itself to itself.
					// XXX A signed shift left likely will be faster.
					ADD(rr, rr, rr);
					ADD(rr, rr, rr);
					ADD(rr, rr, rr);
					ADD(rr, ra, ra);
					return true;
					// Beyond that, there is no savings over mulli with
					// repeated adds except if we can shift. We could
					// optimize 2^x+1 and 2^x-1 in the future.
				default:
					break; // fall through
				}
				// XXX Shifter here. For now, just MULLI, and sigh.
				MULLI(rr, ra, rhsc);
				return true;
			}
		}

		if (isU16(rhsc)) {
			// ppc logical immediate zero-extend the imm16 value
			switch (op) {
			case LIR_ori:
				ORI(rr, ra, rhsc);
				return true;
			case LIR_andi:
				ANDI(rr, ra, rhsc);
				return true;
			case LIR_xori:
				XORI(rr, ra, rhsc);
				return true;
			case LIR_andci:
				// no immediate version for ANDC
				return false;
			}
		}

		// LIR shift ops only use last 5bits of shift const
		switch (op) {

		case LIR_lshi:
			SLWI(rr, ra, rhsc & 0x1F);
			return true;

		case LIR_rshui:
			SRWI(rr, ra, rhsc & 0x1F);
			return true;

		case LIR_rshi:
			SRAWI(rr, ra, rhsc & 0x1F);
			return true;

		case LIR_rori:
			RLWINM(rr, ra, (32 - rhsc) & 0x1F, 0, 31);
			return true;

		case LIR_rcri: {
				int shiftCount = rhsc & 0x1F;
				NanoAssert(shiftCount != 0);

				RLWIMI(rr, R0, (32 - flagCarryIndex - shiftCount) & 0x1F, shiftCount - 1, shiftCount - 1);
				MFXER(R0);
				if (shiftCount > 1)
				{
					OR(rr, rr, R0);
					SRWI(R0, ra, 33 - shiftCount);
				}
				SRWI(rr, ra, shiftCount);
				return true;
			}
		}

		return false;
	}

	void Assembler::asm_arith(LIns *ins) {

		LOpcode op = ins->opcode();
		LIns* lhs = ins->oprnd1();
		LIns* rhs = ins->oprnd2();
		uint32_t flagRequestedMask = ins->flagRequested();

		// Try immediate mode, except if one of the unsupported opcodes.
		// if flags are requested, switch on the non immediate form.
		if (rhs->isImmI() && rhs->canOptimize() && (is_shift_opcode(op) || (flagRequestedMask == 0))) {
			if (asm_arith_imm(ins)) {
				freeResourcesOf(ins);
				return;
			}
		}

		Register rr = prepareResultReg(ins, GpRegs);
		RegisterMask allow = GpRegs & ~rmask(rr);
		Register ra, rb;
		findRegFor2(allow, lhs, ra, allow, rhs, rb);

		int flagCondition   = (flagRequestedMask & (ArithFlag_Zero | ArithFlag_Sign)) == 0 ? 0 : 1;
		int flagOverflow    = (flagRequestedMask & ArithFlag_Overflow) == 0 ? 0 : 1;
		int flagCarry       = (flagRequestedMask & (ArithFlag_Carry | ArithFlag_Borrow)) == 0 ? 0 : 1;

		switch (op) {
			CASE64(LIR_addq:)
			case LIR_addi:
				if (flagCarry == 0)
					ADD_cond(rr, ra, rb, flagCondition, flagOverflow);
				else
					ADDC_cond(rr, ra, rb, flagCondition, flagOverflow);
				break;

			case LIR_addci:
				ADDE_cond(rr, ra, rb, flagCondition, flagOverflow);
				break;

			case LIR_subi:
				if (flagCarry == 0)
					SUBF_cond(rr, rb, ra, flagCondition, flagOverflow);
				else
					SUBFC_cond(rr, rb, ra, flagCondition, flagOverflow);
				break;

			case LIR_subbi:
				SUBFE_cond(rr, rb, ra, flagCondition, flagOverflow);
				//MTXER(R0);
				//XORIS(R0, R0, (1 << 13));
				//MFXER(R0);
				break;

			case LIR_andi:
				NanoAssert((flagOverflow == 0) && (flagCarry == 0));
				AND_cond(rr, ra, rb, flagCondition);
				break;

			case LIR_ori:
				NanoAssert((flagOverflow == 0) && (flagCarry == 0));
				OR_cond(rr, ra, rb, flagCondition);
				break;

			case LIR_xori:
				NanoAssert((flagOverflow == 0) && (flagCarry == 0));
				XOR_cond(rr, ra, rb, flagCondition);
				break;

			case LIR_andci:
				NanoAssert((flagOverflow == 0) && (flagCarry == 0));
				ANDC_cond(rr, ra, rb, flagCondition);
				break;

			case LIR_lshi:
				SLW(rr, ra, rb);
				break;

			case LIR_rshi:
				SRAW(rr, ra, rb);
				break;

			case LIR_rshui:
				SRW(rr, ra, rb);
				break;

			case LIR_rori:
				RLWNM(rr, ra, R0, 0, 31);
				NEG(R0, rb); // 32 - x (hardware will mask bits 0-4)
				break;

			case LIR_rcri:
				NanoAssert(0); // dynamic rcr not allowed
				break;

			case LIR_mullsi:
			case LIR_mullui:
			case LIR_muli:
				NanoAssert((flagOverflow == 0) && (flagCarry == 0));
				MULLW_cond(rr, ra, rb, flagCondition);
				break;

			default:
				verbose_only(outputf("%s", getLirName(op));)
				TODO(asm_arith);
				NanoAssert(0);
				break;
		}

		freeResourcesOf(ins);
	}

	void Assembler::asm_fop(LIns *ins) {
		LOpcode op = ins->opcode();
		LIns* lhs = ins->oprnd1();
		LIns* rhs = ins->oprnd2();
		RegisterMask allow = FpRegs;
		Register rr = prepareResultReg(ins, allow);
		Register ra = ( !lhs->isInReg()
					  ? findSpecificRegFor(lhs, rr)
					  : lhs->getReg() );
		Register rb = rhs==lhs ? ra : findRegFor(rhs, allow&~rmask(ra));
		switch (op) {
			case LIR_addd: FADD(rr, ra, rb); break;
			case LIR_subd: FSUB(rr, ra, rb); break;
			case LIR_muld: FMUL(rr, ra, rb); break;
			case LIR_divd: FDIV(rr, ra, rb); break;
			default:
				verbose_only(outputf("%s", getLirName(op));)
				TODO(asm_fop);
		}
	}

	void Assembler::asm_i2d(LIns *ins) {
		Register r = prepareResultReg(ins, FpRegs);
		Register v = findRegFor(ins->oprnd1(), GpRegs);
		const int d = 16; // natural aligned

		FSUB(r, r, F0);
		LFD(r, d, SP); // scratch area in outgoing linkage area
		STW(R0, d+4, SP);
		XORIS(R0, v, 0x8000);
		LFD(F0, d, SP);
		STW(R0, d+4, SP);
		LIS(R0, 0x8000);
		STW(R0, d, SP);
		LIS(R0, 0x4330);
	}

	void Assembler::asm_ui2d(LIns *ins) {
		Register r = prepareResultReg(ins, FpRegs);
		Register v = findRegFor(ins->oprnd1(), GpRegs);
		const int d = 16;

		FSUB(r, r, F0);
		LFD(F0, d, SP);
		STW(R0, d+4, SP);
		LI(R0, 0);
		LFD(r, d, SP);
		STW(v, d+4, SP);
		STW(R0, d, SP);
		LIS(R0, 0x4330);
	}

    void Assembler::asm_movi2f(LIns *ins) {
        (void)ins;
        NanoAssertMsg(0, "LIR_movi2f not yet supported for this architecture");
    }
    void Assembler::asm_movf2i(LIns *ins) {
        (void)ins;
        NanoAssertMsg(0, "LIR_movf2i not yet supported for this architecture");
    }
    void Assembler::asm_movq2d(LIns *ins) {
        (void)ins;
        NanoAssertMsg(0, "LIR_movq2d not yet supported for this architecture");
    }
	void Assembler::asm_ui2f(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_ui2f not yet supported for this architecture");
	}
	void Assembler::asm_f2d(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_f2d not yet supported for this architecture");
	}
	void Assembler::asm_d2f(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_d2f not yet supported for this architecture");
	}
	void Assembler::asm_immf(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_immf not yet supported for this architecture");
	}
	void Assembler::asm_immf4(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_immf4 not yet supported for this architecture");
	}
	void Assembler::asm_f2f4(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_f2f4 not yet supported for this architecture");
	}
	void Assembler::asm_f4comp(LIns *ins) {
		(void)ins;
		NanoAssertMsg(0, "LIR_f4comp not yet supported for this architecture");
	}
	void
	Assembler::asm_load128(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_load128 not yet supported for this architecture");
	}

	void
	Assembler::asm_store128(LOpcode op, LIns* value, int dr, LIns* base)
	{
		(void)op; (void)value;(void)dr;(void)base;
		NanoAssertMsg(0, "asm_store128 not yet supported for this architecture");
	}

	void Assembler::asm_cvti2f(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvti2f not yet supported for this architecture");
	}

	void Assembler::asm_cvtui2f(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvtui2f not yet supported for this architecture");
	}

	void Assembler::asm_cvti2d(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvti2d not yet supported for this architecture");
	}

	void Assembler::asm_cvtui2d(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvtui2d not yet supported for this architecture");
	}

	void Assembler::asm_cvt2itf(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2itf not yet supported for this architecture");
	}

	void Assembler::asm_cvt2irf(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2irf not yet supported for this architecture");
	}

	void Assembler::asm_cvt2uitf(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2uitf not yet supported for this architecture");
	}

	void Assembler::asm_cvt2uirf(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2uirf not yet supported for this architecture");
	}

	void Assembler::asm_cvt2itd(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2itd not yet supported for this architecture");
	}

	void Assembler::asm_cvt2ird(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2ird not yet supported for this architecture");
	}

	void Assembler::asm_cvt2uitd(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2uitd not yet supported for this architecture");
	}

	void Assembler::asm_cvtf2d(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvtf2d not yet supported for this architecture");
	}

	void Assembler::asm_cvt2uird(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_cvt2uird not yet supported for this architecture");
	}

	void Assembler::asm_movf2d(LIns* ins)
	{
		(void)ins;
		NanoAssertMsg(0, "asm_mov2d not yet supported for this architecture");
	}

	/* RICO not used
	void Assembler::asm_d2i(LIns* ins) {
	// Like SPARC, PPC fctid/fctiw only handles fpr->mem->gpr ultimately.
		LIns *lhs = ins->oprnd1();
		Register rr = prepareResultReg(ins, GpRegs);
		Register ra = findRegFor(lhs, FpRegs);
		const int d = 16; // see asm_i2d for why we're using SP+d, not FP
		LWZ(rr, d+4, SP);
	  #ifdef _PPC970_
		// POWER4 and 5 (and 6/7?) does better if the stfd and lwz aren't
		// in the same dispatch group, including the 970. Two nop()s seems
		// optimal for our workload; more actually degrades runtime.
		NOP();
		NOP();
	  #endif
		STFD(F0, d, SP);
		FCTIW(F0, ra);
	}
	*/

	void Assembler::asm_immd(LIns *ins) {
		Register r = prepareResultReg(ins, FpRegs);

		if (rmask(r) & FpRegs) {
			union {
				double d;
				struct {
					int32_t hi, lo; // Always assuming big-endian in NativePPC.cpp
				} w;
			};
			d = ins->immD();
			LFD(r, 8, SP);
			STW(R0, 12, SP);
			asm_li(R0, w.lo);
			STW(R0, 8, SP);
			asm_li(R0, w.hi);
		}
		else {
			int64_t q = ins->immDasQ();
			if (isS32(q)) {
				asm_li(r, int32_t(q));
				return;
			}
			NanoAssert(0); // must fix, cannot use 64 bit opcode
			//RLDIMI(r,R0,32,0); // or 32,32?
			asm_li(R0, int32_t(q>>32)); // hi bits into R0
			asm_li(r, int32_t(q)); // lo bits into dest reg
		}

		freeResourcesOf(ins);
	}

	void Assembler::br(const void* addr, int link) {

		NanoAssert(addr != NULL);

		// powerpc offsets are based on the address of the branch instruction
		underrunProtect(4); // ensure _nIns is addr of Bx

		//  we want ptr diff's implicit >>2 here
		ptrdiff_t fixedReferenceAddress = reinterpret_cast<ptrdiff_t>(_nIns-1) + _codeAlloc->absoluteOffset();
		ptrdiff_t offset = (reinterpret_cast<ptrdiff_t>(addr) - fixedReferenceAddress) >> 2;
		if (isS24(offset)) {
			Bx(offset, 0, link); // b addr or bl addr
			_relativeOffsetPatchList.push_back(_nIns);
			return;
		}

		uintptr_t destination = reinterpret_cast<uintptr_t>(addr) >> 2;
		if (isS24(destination)) {
			Bx(destination, 1, link); // ba addr or bla addr
			return;
		}

		NanoAssert(0);
		TRAP();
	}

	void Assembler::br_far(const void* addr, int link) {
		NanoAssert(0); // for now, should never been used
	}

	void Assembler::asm_cmov(LIns* ins) {

		NanoAssert(ins->oprnd1()->isCmp());
		NanoAssert((ins->isop(LIR_cmovi) && ins->oprnd2()->isI() && ins->oprnd3()->isI())  ||
			(ins->isop(LIR_cmovd) && ins->oprnd2()->isD() && ins->oprnd3()->isD())  );

		SwapDisable();

		if (ins->opcode() == LIR_cmovd) {
			asm_cmovd(ins);
		} else {
			asm_cmovi_branchless(ins);
		}

		SwapEnable();

	}

	void Assembler::asm_cmovd(LIns* ins)
	{
		// floats. This is a very naive implementation, frankly.
		// However, we can't use fsel here because of LIR's design.
		Register rr;
		Register rf;
		Register rt;

		LIns* condval = ins->oprnd1();
		LIns* iftrue  = ins->oprnd2();
		LIns* iffalse = ins->oprnd3();

		rr = prepareResultReg(ins, FpRegs);
		rf = findRegFor(iffalse, FpRegs & ~rmask(rr));
		// If 'iftrue' isn't in a register, it can be clobbered by 'ins'.
		rt = iftrue->isInReg() ? iftrue->getReg() : rr;

		underrunProtect(20);
		// make sure branch target and branch are on same page and thus near
		NIns *after = _nIns;
		verbose_only(if (_logc->getLcBits() & LC_Native)    outputf("%p:",after);)
			FMR(rr,rf);

		NanoAssert(isS14(after - (_nIns-1)));
		Branches branches;
		memset(&branches, 0, sizeof(branches));
		asm_branch_cond(false, condval, after, false, &branches);

		if (rr != rt)
			FMR(rr, rt);

		freeResourcesOf(ins);
		if (!iftrue->isInReg()) {
			NanoAssert(rt == rr);
			findSpecificRegForUnallocated(iftrue, rr);
		}

		asm_cmp(condval->opcode(), condval->oprnd1(), condval->oprnd2(), CR0);
	}

	void Assembler::asm_cmovi_branchless(LIns* ins) {

		LIns* iftrue  = ins->oprnd2();
		LIns* iffalse = ins->oprnd3();
		LIns* condval = ins->oprnd1();

		bool isImm0 = (condval->oprnd2()->canOptimize() && condval->oprnd2()->isImmI() && (condval->oprnd2()->immI() == 0));

		// condition operand must be set to live
		setResultLive(condval->oprnd1());
		setResultLive(condval->oprnd2());

		RegisterMask regMask = GpRegs;
		Register rr, rf, rt, ra, rb, rsel;

		rr = prepareResultReg(ins, regMask);
		regMask &= ~rmask(rr);

		if (!iffalse->isInReg()) {
			rf = findRegFor(iffalse, regMask);
			regMask &= ~rmask(rf);
		} else {
			rf = iffalse->getReg();
			NanoAssert(rmask(rf) & GpRegs);
		}

		if (!iftrue->isInReg()) {
			rt = findRegFor(iftrue, regMask);
			regMask &= ~rmask(rt);
		} else {
			rt = iftrue->getReg();
			NanoAssert(rmask(rt) & GpRegs);
		}

		if (!condval->oprnd1()->isInReg()) {
			ra = findRegFor(condval->oprnd1(), regMask);
			regMask &= ~rmask(ra);
		} else {
			ra = condval->oprnd1()->getReg();
			NanoAssert(rmask(ra) & GpRegs);
		}

		if (!isImm0) {
			if (!condval->oprnd2()->isInReg()) {
				rb = findRegFor(condval->oprnd2(), regMask);
				regMask &= ~rmask(rb);
			} else {
				rb = condval->oprnd2()->getReg();
				NanoAssert(rmask(rb) & GpRegs);
			}
		} 

		rsel = allocTempReg(regMask);

		underrunProtect(32);

		OR(rr, rr, R0);
		AND(rr, rt, rsel);
		ANDC(R0, rf, rsel);

		if (isImm0)
			asm_cmovi_createnegmask_imm0(condval->opcode(), ra, rsel);
		else
			asm_cmovi_createnegmask(condval->opcode(), ra, rb, rsel);

		freeResourcesOf(ins);
	}

	void Assembler::asm_cmovi_createnegmask(LOpcode op, Register ra, Register rb, Register r)
	{
	/*
	create a neg mask from a comparison operator, section D.2
	https://www-01.ibm.com/chips/techlib/techlib.nsf/techdocs/852569B20050FF7785256996007558C6/$file/cwg.pdf
	*/
		switch (op)
		{
			case LIR_eqi:
			{
				SUBFE(r, R0, R0);
				ADDIC(r, R0, -1);
				SUBF(R0, rb, ra);
				break;
			}

			case LIR_nei:
			{
				SUBFE(r, r, r);
				SUBFIC(r, R0, 0);
				SUBF(R0, rb, ra);
				break;
			}

			case LIR_gei: // signed greater than or equal
			{
				Register rtemp = ra;
				ra = rb;
				rb = rtemp;
				// fall through
			}

			case LIR_lei: // less than or equal
			{
				SUBFE(r, r, r);
				ADDC(R0, r, R0);
				SUBF(r, ra, rb);
				XORIS(R0, ra, 0x8000);
				break;
			}

			case LIR_gti: // signed greater-than
			{
				Register rtemp = ra;
				ra = rb;
				rb = rtemp;
				// fall through
			}

			case LIR_lti: // signed less than
			{
				SUBFE(r, R0, r);
				SRWI(R0, ra, 31);
				SRWI(r, rb, 31);
				SUBFC(R0, rb, ra);
				break;
			}

			case LIR_geui:
			{
				Register rtemp = ra;
				ra = rb;
				rb = rtemp;
				// fall through
			}

			case LIR_leui:
			{
				SUBF(r, R0, ra);
				ADDZE(R0, ra);
				SUBFC(r, ra, rb);
				break;
			}

			case LIR_gtui:
			{
				Register rtemp = ra;
				ra = rb;
				rb = rtemp;
				// fall through
			}

			case LIR_ltui:
			{
				SUBFE(r, r, r);
				SUBFC(R0, rb, ra);
				break;
			}

			default:
				NanoAssert(0); // only "i" opcode are allowed
		}
	}

	void Assembler::asm_cmovi_createnegmask_imm0(LOpcode op, Register ra, Register r)
	{
		/*
		create a neg mask from a comparison operator, section D.2
		https://www-01.ibm.com/chips/techlib/techlib.nsf/techdocs/852569B20050FF7785256996007558C6/$file/cwg.pdf
		*/

		switch (op)
		{
			case LIR_geui:
				ADDI(r, R0, 0xFFFF); // always true => 0xFFFFFFFF
				break;

			case LIR_leui:
			case LIR_eqi:
				SUBFE(r, r, r);
				ADDIC(R0, ra, -1);
				break;

			case LIR_gtui:
			case LIR_nei:
				SUBFE(r, r, r);
				SUBFIC(R0, ra, 0);
				break;

			case LIR_gei: // signed greater than or equal
				ADDI(r, r, -1);
				SRWI(r, ra, 31);
				break;

			case LIR_lei: // less than or equal
				SUBFZE(r, R0);
				SRWI(R0, ra, 31);
				ADDIC(R0, ra, -1);
				break;

			case LIR_gti: // signed greater-than
				ADDME(r, R0);
				SRWI(R0, ra, 31);
				SUBFIC(r, ra, 0);
				break;

			case LIR_lti: // signed less than
				SRAWI(r, ra, 31);
				break;

			case LIR_ltui:
				LI(r, 0);
				break;

			default:
				NanoAssert(0); // only "i" opcode are allowed
		}
	}


	void Assembler::asm_neg_not(LIns *ins) {
		Register rr = prepareResultReg(ins, GpRegs);
		LIns* lhs = ins->oprnd1();
		// We can clobber the result register for "free."
		Register ra = ( !lhs->isInReg()
					  ? rr
					  : lhs->getReg() );

		NanoAssert((ins->flagRequested() & (ArithFlag_Carry | ArithFlag_Borrow | ArithFlag_Overflow)) == 0);
		int condition = (ins->flagRequested() != 0) ? 1 : 0;

		if (ins->isop(LIR_negi)) {
			NEG_cond(rr, ra, condition);
		} else {
			NOT_cond(rr, ra, condition);
		}

		freeResourcesOf(ins);

		if (rr == ra)
			findSpecificRegFor(lhs, ra);
	}

	void Assembler::nCreate() {
	}

	void Assembler::nDestroy() {
	}

	void Assembler::nBeginAssembly(nanojit::Allocator*) {
	}

	void Assembler::nEndAssembly(nanojit::Allocator*) {
	}

	// Increment the 32-bit profiling counter at pCtr, without
	// changing any registers.
	verbose_only(
	void Assembler::asm_inc_m32(uint32_t* pCtr)
	{
		*pCtr = *pCtr + 1;
	}
	)

	static void emit_trap(NIns *ins) {
		NanoAssert(0);
		*ins = PPC_tw | (0x1F << 21); // trap always
	}

	void Assembler::nPatchBranch(NIns *branch, const NIns *target) {
		// ppc relative offsets are based on the addr of the branch instruction
		ptrdiff_t bd = target - branch;
		NIns branchOp = *branch;
		const uint32_t opcodeMask = 63UL << 26; // bits [0-5]
		

		// bitmasks courtesy http://www.mactech.com/articles/mactech/Vol.10/10.09/PowerPCAssembly/
		if ((branchOp & opcodeMask) == PPC_b) {
			// unconditional, 24bit offset.Whoever generated the unpatched jump
			// must have known the final size would fit in 24bits!  otherwise the
			// jump would be (lis,ori,mtctr,bctr) and we'd be patching the lis,ori.
			NanoAssert(isS24(bd)); // 26 bits allowed in offset, but sizeof(NIns) is 4
			if (!isS24(bd)) {
				emit_trap(branch);
				return;
			}
			
			const uint32_t offsetMask = 0xffffff << 2;
			branchOp &= ~offsetMask;
			branchOp |= (bd << 2) & offsetMask;
			*branch = branchOp;
			return;

		} else if ((branchOp & opcodeMask) == PPC_bc) {
			// conditional, 14bit offset. Whoever generated the unpatched jump
			// must have known the final size would fit in 14bits! otherwise the
			// jump would be (lis,ori,mtctr,bcctr) and we'd be patching the lis,ori below.
			NanoAssert(isS14(bd));
			if (!isS14(bd)) {
				emit_trap(branch);
				return;
			}
			const uint32_t offsetMask = 0x3fff << 2;
			branchOp &= ~offsetMask;
			branchOp |= (bd << 2) & offsetMask;
			*branch = branchOp;
			return;

		} else {
			emit_trap(branch);
		}
	}

	void Assembler::nPatchRelativeOffset(const NIns *originalAddress, NIns *opcodeAddress) {
		const uint32_t opcodeMask = (63UL << 26) | 0x03; // bits [0-5]
		const uint32_t offsetMask = 0xffffff << 2;

		uint32_t opcode = *opcodeAddress;
		NanoUnusedVariable(opcodeMask);
		NanoAssert((opcode & opcodeMask) == (PPC_b | 0x01)); // BL
		int32_t offset = (int32_t(opcode & offsetMask) << 6) >> 6; // sign extend offset
		intptr_t baseAddress = reinterpret_cast<intptr_t>(originalAddress);
		intptr_t destinationAddress = baseAddress + offset;
		const NIns* target = reinterpret_cast<const NIns*>(destinationAddress);
		ptrdiff_t bd = target - opcodeAddress;

		if (!isS24(bd)) {
			emit_trap(opcodeAddress);
			return;
		}

		opcode &= ~offsetMask;
		opcode |= (bd << 2) & offsetMask;
		*opcodeAddress = opcode;
		
		// ppc relative offsets are based on the addr of the branch instruction
	}


	void Assembler::asm_insert_random_nop() {
		NanoAssert(0); // not supported
	}

	void Assembler::asm_label() {
#if defined(NANOJIT_PPC_ENABLE_SWAP)
		// disable swapping for the next instruction; it's a branch point.
		_lastOpcode.reg1 = NoSwap;
#endif
	}

	void Assembler::asm_movi(LIns* ins) {

		LIns* oprnd1 = ins->oprnd1();
		prepareResultReg(ins, GpRegs);

		Register rr = ins->getReg();
		Register ra = oprnd1->isInReg() ? oprnd1->getReg() : rr;

		if (rr != ra)
			MR(rr, ra);

		freeResourcesOf(ins);

		if (!oprnd1->isInReg())
			_allocator.allocSpecificReg(oprnd1, ra);
	}

	void Assembler::asm_movf(LIns* ins) {
		LIns* oprnd1 = ins->oprnd1();
		prepareResultReg(ins, FpRegs);

		Register rr = ins->getReg();
		Register ra = oprnd1->isInReg() ? oprnd1->getReg() : rr;

		if (rr != ra)
			asm_nongp_copy(rr, ra);

		freeResourcesOf(ins);

		if (!oprnd1->isInReg())
			_allocator.allocSpecificReg(oprnd1, ra);

	}

	void Assembler::asm_movd(LIns* ins) {
		LIns* oprnd1 = ins->oprnd1();
		prepareResultReg(ins, FpRegs);

		Register rr = ins->getReg();
		Register ra = oprnd1->isInReg() ? oprnd1->getReg() : rr;

		if (rr != ra)
			asm_nongp_copy(rr, ra);

		freeResourcesOf(ins);

		if (!oprnd1->isInReg())
			_allocator.allocSpecificReg(oprnd1, ra);
	}

	void Assembler::asm_movf4(LIns* ins) {
		NanoAssert(0); // no vector support
	}

	bool Assembler::is_shift_opcode(LOpcode op) {
		return (op == LIR_lshi) || (op == LIR_rshi) || (op == LIR_rshui) || (op == LIR_rori) || (op == LIR_rcri);
	}

	void Assembler::asm_shift_carry(Register r, LIns* ins) {

		NanoAssert(is_shift_opcode(ins->opcode()));

		Register rr = ins->getReg();
		LIns* oprnd1 = ins->oprnd1();
		LIns* oprnd2 = ins->oprnd2();

		// these operands might not have been marked as live yet
		setResultLive(oprnd1);
		setResultLive(oprnd2);

		RegisterMask regMask = GpRegs & ~(rmask(r) | rmask(rr));

		if (oprnd2->isImmI() && oprnd2->canOptimize()) {
			int immI = oprnd2->immI();
			NanoAssert(immI != 0);

			Register ra = findRegFor(oprnd1, regMask);

			switch (ins->opcode()) {
				case LIR_lshi:
					RLWINM(r, ra, immI & 0x1F, 31, 31);
					break;

				case LIR_rshi:
				case LIR_rshui:
				case LIR_rori:
				case LIR_rcri:
					RLWINM(r, ra, (33 - immI) & 0x1F, 31, 31);
					break;

				default:
					NanoAssert(0); // invalid opcode
					break;
			}

			return;
		}


		SwapDisable();

		Register ra = findRegFor(oprnd1, regMask);
		Register rb = findRegFor(oprnd2, regMask & ~(rmask(ra)));

		switch (ins->opcode()) {
			case LIR_lshi:
				RLWNM(r, ra, rb, 31, 31);
				break;

			case LIR_rshi:
			case LIR_rshui:
			case LIR_rori:
			case LIR_rcri:
				RLWNM(r, ra, R0, 31, 31);
				SUBFIC(R0, rb, 33); // 33 - shift count, will dirty carry flag
				break;

			default:
				NanoAssert(0); // invalid opcode
				break;
		}


		SwapEnable();
	}


	void Assembler::asm_save_flag(LIns* ins) {

		SwapDisable();

		LIns*       oprnd1  = ins->oprnd1();

		Register    r = prepareResultReg(ins, GpRegs);

		if (oprnd1->isImmAny() && oprnd1->canOptimize()) {
			asm_save_flag_imm(ins);
			return;
		}

		findRegFor(oprnd1, GpRegs & ~rmask(r));

		LOpcode op = ins->opcode();
		switch (op) {
			case LIR_getflgz:
				RLWINM(r, r, 32 - flagZeroIndex, 31, 31);
				MFCR(r);
				break;

			case LIR_getflgnz:
				RLWINM(r, r, 32 - flagZeroIndex, 31, 31);
				NOT(r, r);
				MFCR(r);
				break;

			case LIR_getflgs:
				RLWINM(r, r, 32 - flagSignIndex, 31, 31);
				MFCR(r);
				break;

			case LIR_getflgns:
				RLWINM(r, r, 32 - flagSignIndex, 31, 31);
				NOT(r, r);
				MFCR(r);
				break;

			case LIR_getflgc: {
				NanoAssert((oprnd1->opcode() != LIR_subbi) && (oprnd1->opcode() != LIR_subi));

				LOpcode oprnd1Op = oprnd1->opcode();
				if (is_shift_opcode(oprnd1Op)) {
					asm_shift_carry(r, oprnd1);
				} else {
					RLWINM(r, r, 32 - flagCarryIndex, 31, 31);
					MFXER(r);
				}
				break;
			}

			case LIR_getflgb: {
				NanoAssert((oprnd1->opcode() == LIR_subbi) || (oprnd1->opcode() == LIR_subi));

				RLWINM(r, r, 32 - flagCarryIndex, 31, 31);
				NOT(r, r);
				MFXER(r);
				break;
			}

			case LIR_getflgnc: {
				NanoAssert((oprnd1->opcode() != LIR_subbi) && (oprnd1->opcode() != LIR_subi));

				LOpcode oprnd1Op = oprnd1->opcode();
				if (is_shift_opcode(oprnd1Op)) {
					NOT(r, r);
					asm_shift_carry(r, oprnd1);
				} else {
					RLWINM(r, r, 32 - flagCarryIndex, 31, 31);
					NOT(r, r);
					MFXER(r);
				}
				break;
			}

			case LIR_getflgnb: {
				NanoAssert((oprnd1->opcode() == LIR_subbi) || (oprnd1->opcode() == LIR_subi));

				RLWINM(r, r, 32 - flagCarryIndex, 31, 31);
				MFXER(r);
				break;
			}

			case LIR_getflgv:
				RLWINM(r, r, 32 - flagOverflowIndex, 31, 31);
				MFXER(r);
				break;

			case LIR_getflgnv:
				RLWINM(r, r, 32 - flagOverflowIndex, 31, 31);
				NOT(r, r);
				MFXER(r);
				break;

			default:
				NanoAssert(0); // not implemented
				break;
		}

		freeResourcesOf(ins);

		SwapEnable();
	}

	void Assembler::asm_save_flag_imm(LIns* ins) {
		LOpcode     op  = ins->opcode();
		Register    rr  = prepareResultReg(ins, GpRegs);
		LIns*       a   = ins->oprnd1();

		NanoAssert((ins->flagRequested() & (ArithFlag_Overflow | ArithFlag_Carry | ArithFlag_Borrow)) == 0);
		NanoAssert(a->canOptimize());

		if (a->isImmI())
		{
			uint32_t iImm = a->immI();

			switch (op)
			{
			case LIR_getflgz:       LI(rr, iImm == 0 ? 1 : 0);      break;
			case LIR_getflgs:       LI(rr, iImm >> 31);             break;
			case LIR_getflgnz:      LI(rr, iImm != 0 ? 1 : 0);      break;
			case LIR_getflgns:      LI(rr, 1 - (iImm >> 31));       break;
			default:                NanoAssert(0);                  break;
			}
		}

		freeResourcesOf(ins);
	}

	void Assembler::asm_restore_flag(LIns* ins) {

		SwapDisable();

		Register r = R0; // R0 is always available for temporary storage
		Register s = findRegFor(ins->oprnd1(), GpRegs);

		unsigned int iCarryMaskIndex = 31 - flagCarryIndex;

		switch (ins->opcode())
		{

		case LIR_setflgnb:
		case LIR_setflgc: {
			MTXER(r);
			RLWIMI(r, s, flagCarryIndex, iCarryMaskIndex, iCarryMaskIndex);
			MFXER(r);
			break;
		}

		case LIR_setflgb:
		case LIR_setflgnc: {
			Register t = allocTempReg(GpRegs);
			MTXER(r);
			RLWIMI(r, t, flagCarryIndex, iCarryMaskIndex, iCarryMaskIndex);
			NOT(t, s);
			MFXER(r);
			break;
		}

		default:
			NanoAssert(0); // not implemented
		}


		SwapEnable();
	}

	void Assembler::asm_mull(LIns* ins) {
		asm_arith(ins);
	}

	void Assembler::asm_mullhi(LIns* ins) {
		LOpcode op = ins->opcode();
		LIns* mull = ins->oprnd1();

		mull->oprnd1()->setResultLive();
		mull->oprnd2()->setResultLive();

		NanoAssert((mull->opcode() == LIR_mullui) || (mull->opcode() == LIR_mullsi));

		Register rr = prepareResultReg(ins, GpRegs);
		RegisterMask allow = GpRegs & ~rmask(rr);
		Register ra, rb;
		findRegFor2(
			allow, mull->oprnd1(), ra,
			allow, mull->oprnd2(), rb);

		int flagCondition = (ins->flagRequested() & (ArithFlag_Zero | ArithFlag_Sign)) == 0 ? 0 : 1;
		NanoAssert((ins->flagRequested() & (ArithFlag_Overflow | ArithFlag_Carry | ArithFlag_Borrow)) == 0);

		switch (op)
		{
			case LIR_mullshii:
				MULHW_cond(rr, ra, rb, flagCondition);
				break;

			case LIR_mulluhii:
				MULHWU_cond(rr, ra, rb, flagCondition);
				break;

			default:
				NanoAssert(0); // invalid / unknow opcode
				break;
		}

		freeResourcesOf(ins);
	}

	void Assembler::asm_recip_sqrt(LIns* ins) {
		NanoAssert(0); // not implemented
	}

	void Assembler::asm_force_gen_flags(LIns* ins) {
		NanoAssert(ins->isI());
		Register r = findRegFor(ins, GpRegs);
		CMPWI(CR0, r, 0);
	}

	void Assembler::asm_store_to_stack(LIns* ins, int d, RegisterMask allow) {
		NanoAssert(0); // should never be called, parameters are always passed by registers
	}

	// regalloc
	const Register RegAlloc::retRegs[] = { R3, R4 }; // high=R3, low=R4
	const Register RegAlloc::argRegs[] = { R3, R4, R5, R6, R7, R8, R9, R10 };

	const Register RegAlloc::savedRegs[] = {
		R13,
		R14, R15, R16, R17, R18, R19, R20, R21, R22,
		R23, R24, R25, R26, R27, R28, R29, R30
	};


	Register RegAlloc::nRegisterAllocFromSet(RegisterMask set) {
		uint32_t i;
		// note, deliberate truncation of 64->32 bits
		if (set & 0xffffffff) {
			i = 31 - countLeadingZeroes(uint32_t(set)); // gp reg
		} else {
			i = 32 + 31 - countLeadingZeroes(uint32_t(set>>32)); // fp reg
		}
		Register r = { i };
		return r;
	}

	RegisterMask RegAlloc::nInitManagedRegisters(const Config* config) {
		NanoUnusedVariable(config);
		return SavedRegs | 0x1ff8 /* R3-12 */ | 0x3ffe00000000LL /* F1-13 */;
	}

	// Init per-opcode register hint table. Defaults to no hints for all instructions
	// (initialized to 0 )
	void RegAlloc::nInitHint() {
		VMPI_memset(_hint, 0, sizeof(RegisterMask)*LIR_sentinel );
		_hint[LIR_calli]  = rmask(R3);
		_hint[LIR_calld]  = rmask(F1);
		_hint[LIR_paramp] = PREFER_SPECIAL;
	}

	RegisterMask RegAlloc::nHint(LIns* ins) {
		NanoAssert(_hint[ins->opcode()] == PREFER_SPECIAL);
		NanoAssert(ins->isLInsP());

		RegisterMask prefer = 0;
		Register reg = ins->paramReg();
		if (reg != UnspecifiedReg)
			prefer = rmask(reg);

		return prefer;
	}

	RegisterMask RegAlloc::nRegCopyCandidates(Register r, RegisterMask allow) {
		// PPC doesn't support any GPR<->FPR moves
		if(rmask(r) & GpRegs)
			return allow & GpRegs;
		if(rmask(r) & FpRegs)
			return allow & FpRegs;
		NanoAssert(false); // How did we get here?
		return 0;
	}

	bool RegAlloc::canRemat(LIns* ins) {
		return (ins->isImmI() && ins->canOptimize());
	}


} // namespace nanojit

#endif // FEATURE_NANOJIT && NANOJIT_PPC


