/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */


#ifndef __nanojit_Assembler__
#define __nanojit_Assembler__

namespace nanojit
{
	/**
	 * Some notes on this Assembler (Emitter).
	 *
	 *   The class RegAlloc is essentially the register allocator from MIR
	 *
	 *   The Assembler class parses the LIR instructions starting at any point and converts
	 *   them to machine code.  It does the translation using expression trees which are simply
	 *   LIR instructions in the stream that have side-effects.  Any other instruction in the
	 *   stream is simply ignored.
	 *   This approach is interesting in that dead code elimination occurs for 'free', strength
	 *   reduction occurs fairly naturally, along with some other optimizations.
	 *
	 *   A negative is that we require state as we 'push' and 'pop' nodes along the tree.
	 *   Also, this is most easily performed using recursion which may not be desirable in
	 *   the mobile environment.
	 *
	 */

	#ifndef AVMPLUS_ALIGN16
		#ifdef _MSC_VER
			#define AVMPLUS_ALIGN16(type) __declspec(align(16)) type
		#else
			#define AVMPLUS_ALIGN16(type) type __attribute__ ((aligned (16)))
		#endif
	#endif

	class Noise
	{
		public:
			virtual ~Noise() {}

			// produce a random number from 0-maxValue for the JIT to use in attack mitigation
			virtual uint32_t getValue(uint32_t maxValue) = 0;
	};

	// error codes
	enum AssemblerError
	{
		 None = 0
		,StackFull
		,UnknownBranch
		,BranchTooFar
	};

	struct CompilationStats
	{
		size_t      reconcileBackwardMoveToRegCount;
		size_t      reconcileForwardMoveToRegCount;
		size_t      reconcileForceSpillCount;
		size_t      reconcileForceRestoreCount;
		size_t      spillCount;
		size_t      restoreCount;
	};

	typedef SeqBuilder<NIns*> NInsList;
	typedef HashMap<NIns*, LIns*, 16> NInsMap;
	typedef PodVector<Pair<NIns*, LIns*> > BranchPatchList;
	typedef PodVector<Pair<NIns*, const NIns*> > RelocationList;
	typedef PodVector<const void*> EntryPointList;
	typedef PodVector<const NIns*> RelativeOffsetPatchList;

#if NANOJIT_USES_IMMD_POOL
	typedef HashMap<uint64_t, uint64_t*> ImmDPoolMap;
#endif
#if NANOJIT_USES_IMMF4_POOL
	#if defined(_WIN32) && !defined(NANOJIT_X64) //AVMSYSTEM_32BIT
		// The Win32 compiler for i386 provides only 8-byte alignment for the stack,
		// thus the required 16-byte alignment for float4_t cannot be guaranteed.
		// Unfortunately, HashMap members pass the key as a value parameter in several
		// places, causing compilation errors.  As a workaround, we encapsulate the
		// key value in a "proxy" class whose instances need not themselves be aligned,
		// but are coercible to properly-aligned copy of the underlying float4_t value.
		class float4_key {
		private:
			float _val[4];
		public:
			float4_key() {} // for alignof macro, don't call.
			float4_key(const float4_t& p) { _val[0] = f4_x(p); _val[1] = f4_y(p); _val[2] = f4_z(p); _val[3] = f4_w(p); }
			operator float4_t() const { float4_t f4value = { _val[0], _val[1], _val[2], _val[3] }; return f4value; }
			bool keyEquals(const float4_key& other) const {  return VMPI_memcmp(_val, other._val, sizeof(_val)) == 0; }
		};

		template <> struct DefaultKeysEqual<float4_key> {
			static bool keysEqual(const float4_key& x, const float4_key& y) {
				return x.keyEquals(y);
			}
		};

		typedef HashMap<float4_key, float4_t*, 16> ImmF4PoolMap;
	#else
		// Other supported compilers provide for the necessary stack alignment.
		typedef HashMap<float4_t, float4_t*, 16> ImmF4PoolMap;
	#endif
#endif //NANOJIT_USES_IMMF4_POOL

	/**
-    * Some architectures (i386, X64) can emit two branches that need patching
-    * in some situations. This is returned by asm_branch() implementations
-    * with 0, 1 or 2 of these fields set to a non-NULL value. (If only 1 is set, 
-    * it must be patch1, not patch2.)
-    */
	class Branches
	{
		static const unsigned int MAX_BRANCH = 3;

		NIns* branch[MAX_BRANCH];
		const NIns* destination[MAX_BRANCH];
		size_t count;

	public:

		inline Branches()
			: count(0) {
		}

		inline void addBranch(NIns* b, const NIns* d) {
			NanoAssert(count < MAX_BRANCH);
			branch[count] = b;
			destination[count] = d;
			count++;
		}

		inline size_t getCount() const {
			return count;
		}

		inline NIns* getBranch(size_t branchIndex) {
			NanoAssert(branchIndex < count);
			return branch[branchIndex];
		}

		inline const NIns* getDestination(size_t branchIndex) {
			NanoAssert(branchIndex < count);
			return destination[branchIndex];
		}
	};

	/**
	 * Information about the activation record for the method is built up
	 * as we generate machine code.  As part of the prologue, we issue
	 * a stack adjustment instruction and then later patch the adjustment
	 * value.  Temporary values can be placed into the AR as method calls
	 * are issued.
	 */
	class Assembler
	{
#ifdef NANOJIT_VERBOSE
		public:
			// Buffer for holding text as we generate it in reverse order.
			StringList* _outputCache;

			// Outputs the format string and 'outlineEOL', and resets
			// 'outline' and 'outlineEOL'.
			void outputf(const char* format, ...);

		private:
			static int disassembler_printf(void* arg, const char* fmt, ...);
			static void* disassembler_event(void* cookie, const char* event, void* arg);

			// Log controller object.  Contains what-stuff-should-we-print
			// bits, and a sink function for debug printing.
			LogControl* _logc;

			// instruction printer
			LInsPrinter* _printer;

			// Buffer used in most of the output function.  It must big enough
			// to hold both the output line and the 'outlineEOL' buffer, which
			// is concatenated onto 'outline' just before it is printed.
			static char  outline[8192];
			// Buffer used to hold extra text to be printed at the end of some
			// lines.
			static char  outlineEOL[512];

			// Outputs 'outline' and 'outlineEOL', and resets them both.
			// Output goes to '_outputCache' if it's non-NULL, or is printed
			// directly via '_logc'.
			void output();

			void maybe_disassemble();
			void disassemble(void* from, void* to);

			// Sets 'outlineEOL'.
			void setOutputForEOL(const char* format, ...);

			void printRegState();
			void printActivationState();
#endif // NANOJIT_VERBOSE

		public:
			Assembler(CodeAlloc* codeAlloc, Allocator* internalAlloc,
					  const Config& config
					  verbose_only(, LogControl* logc, LInsPrinter* printer = NULL));
			~Assembler();

			void								compile(LirChainWriter* writer, unsigned int arGroupCount);

			const Config&						getConfig() const;

			void								reset(unsigned int ar);
			
			AssemblerError						getError() const;
			bool								hasError() const;

#ifdef NANOJIT_VERBOSE
			const CompilationStats*				getCompilationStats() const;
#endif
			inline CodeBlockList				getCodeBlockList() const;
			// note: entry points are in reverse order
			inline const EntryPointList&			getEntryPointList() const;
			inline const RelativeOffsetPatchList&	getRelativeOffsetPatchList() const;

			inline const RelocationList&		getBranchRelocationList() const;

			ptrdiff_t							computeLinearOffset(const NIns* ins, CodeBlock const** ownerBlock) const;
			bool								flattenCodeBlockList(void* buffer, size_t bufferSize) const;
			bool								computeEntryPointLinearOffsetList(ptrdiff_t* linearOffsetList, size_t linearOffsetListSize) const;

			void								clearCompilationResults();

			static inline void					patchRelativeOffset(const NIns *originalAddress, NIns *opcode) { nPatchRelativeOffset(originalAddress, opcode); }

private:
			void        beginAssembly(unsigned int arGroupCount);
			void        assemble(LirFilter* reader);
			void        endAssembly();

			void        setNoiseGenerator(Noise* noise)  { _noise = noise; } // used for attack mitigation; setting to 0 disables all mitigations

			void        releaseRegisters();
			void        setError(AssemblerError e) { _err = e; }
			void        cleanupAfterError();
			void        clearNInsPtrs();

			debug_only ( void       pageValidate(); )

			// support calling out from a fragment ; used to debug the jit
			debug_only( void        resourceConsistencyCheck(); )
			debug_only( void        registerConsistencyCheck(); )

private:
			void        gen(LirFilter* toCompile);
			NIns*       genEntryPoint(RegisterMask saveMask);
			NIns*       genExitPoint(RegisterMask restoreMask);
			void        genPrologue();

			uint32_t    arReserve(LIns* ins);
			void        arRelease(LIns* ins);
			void        arReset(unsigned int arGroupCount);

			void        evictAllActiveRegs() {
				// The evicted set will be be intersected with activeSet(),
				// so use an all-1s mask to avoid an extra load or call.
				evictSomeActiveRegs(~RegisterMask(0));
			}
			void        evictSomeActiveRegs(RegisterMask regs);

			Register    getBaseReg(LIns *ins, int &d, RegisterMask allow);
			void        getBaseReg2(RegisterMask allowValue, LIns* value, Register& rv,
									RegisterMask allowBase, LIns* base, Register& rb, int &d);
#if NANOJIT_USES_IMMD_POOL
			const uint64_t* findImmDFromPool(uint64_t q);
#endif
#if NANOJIT_USES_IMMF4_POOL
			const float4_t* findImmF4FromPool(const float4_t& q);
#endif

			int         findMemFor(LIns* ins);
			void        reallocateArEntry(LIns* ins, uint32_t entry);
			Register    findRegFor(LIns* ins, RegisterMask allow);
			void        findRegFor2(RegisterMask allowa, LIns* ia, Register &ra,
									RegisterMask allowb, LIns *ib, Register &rb);
			Register    findSpecificRegFor(LIns* ins, Register r);
			Register    findSpecificRegForUnallocated(LIns* ins, Register r);
			Register    prepareResultReg(LIns *ins, RegisterMask allow);
			void        freeResourcesOf(LIns *ins);
			void        evictIfActive(Register r);
			void        evict(LIns* vic);

			void        setBlockMargins(CodeBlock* block, NIns* start, NIns* end);

			void        codeAlloc();
			void        finalizeCodeAlloc();


			Allocator*					_internalAlloc;     // for items with same lifetime as this Assembler
			CodeAlloc*					_codeAlloc;         // for code we generate

			BranchPatchList				_branchPatchList;
			RelocationList				_branchRelocationList;
			EntryPointList				_entryPointList;
			RelativeOffsetPatchList		_relativeOffsetPatchList;

			Noise*              _noise;             // object to generate random noise used when hardening enabled.

		#if NANOJIT_USES_IMMD_POOL
			ImmDPoolMap         _immDPool;
		#endif
		#if NANOJIT_USES_IMMF4_POOL
			ImmF4PoolMap        _immF4Pool;
		#endif

private:
			CodeBlockList	_codeBlockList;         // used blocks of code.
			CodeBlock*		currentCodeBlock;
			NIns*			_nIns;                  // current instruction in current normal code chunk
			NIns*			_nLastIns;

		#ifdef NANOJIT_VERBOSE
			NIns*       _nInsAfter;             // next instruction (ascending) in current normal/exit code chunk (for verbose output)
			size_t      codeBytes;              // bytes allocated in normal code chunks

			CompilationStats	compilationStats;
		#endif

			AssemblerError   _err;           // 0 = means assemble() appears ok, otherwise it failed

			// Holds the current instruction during gen().
			LIns*       currIns;

			AR          _activation;
			RegAlloc    _allocator;

			const Config& _config;


			unsigned int	_nReservedNInsCount;

			void		underrunProtect(size_t count);
			bool		reserveNInsSlotAtCodeStart(unsigned int count);
			void		freeReservedNInsSlot(unsigned int count);


			static unsigned int computeInitialReservedNInsCount(NIns* curIns, CodeBlock* codeBlock);

			void		nativePageReset();
			void		nativePageSetup();
			void		nativePageSplit();

			CodeBlock*	findOwnerBlock(const NIns* ins);
			

			void        asm_cond(LIns* ins);
			void        asm_condd(LIns* ins);
			void        asm_spill(Register rr, int d, int8_t nWords);

			verbose_only( void asm_inc_m32(uint32_t*); )
			void        asm_mmq(Register rd, int dd, Register rs, int ds);
			void        asm_jmp(LIns* ins);
			void        asm_jcc(LIns* ins);
			void        asm_entry(LIns* ins);
			void        asm_exit(LIns* ins);
			void        asm_store32(LOpcode op, LIns *val, int d, LIns *base);
			void        asm_store64(LOpcode op, LIns *val, int d, LIns *base);

			void        asm_save_flag(LIns* ins);
			void        asm_restore_flag(LIns* ins);

			// WARNING: the implementation of asm_restore() should emit fast code
			// to rematerialize instructions where canRemat() returns true.
			// Otherwise, register allocation decisions will be suboptimal.
			void        asm_restore(LIns*, Register);

			bool        asm_maybe_spill(LIns* ins, bool pop);

#ifdef NANOJIT_IA32
	#define ASM_SPILL_RESTARGS  , bool pop
#else
	#define ASM_SPILL_RESTARGS  
#endif
			void        asm_spill(const LIns* insToSpill ASM_SPILL_RESTARGS);

#ifdef NANOJIT_64BIT
			void        asm_immq(LIns* ins);
#endif
			void        asm_immd(LIns* ins);
			void        asm_arith(LIns* ins);
			void        asm_neg_not(LIns* ins);
			void        asm_mull(LIns* ins);
			void        asm_mullhi(LIns* ins);
			void        asm_load32(LIns* ins);
			void        asm_load64(LIns* ins);
			void        asm_cmov(LIns* ins);
			void        asm_param(LIns* ins);
			void        asm_immi(LIns* ins);
			void        asm_neg_abs(LIns* ins); // fpu neg, abs
			void        asm_recip_sqrt(LIns* ins);  // fpu recip, rsqrt, sqrt
			void        asm_fop(LIns* ins);     // fpu add, sub, mul, div
			void        asm_i2d(LIns* ins);
			void        asm_ui2d(LIns* ins);
			
#ifdef NANOJIT_64BIT
			void        asm_q2i(LIns* ins);
			void        asm_q2d(LIns* ins);
			void        asm_d2q(LIns* ins);
			void        asm_ui2uq(LIns *ins);
			void        asm_dasq(LIns *ins);
			void        asm_qasd(LIns *ins);
#endif
			void        asm_mmi(Register rd, int dd, Register rs, int ds);
			void        asm_store128(LOpcode op, LIns *val, int d, LIns *base);
			void        asm_load128(LIns* ins);
			void        asm_immf(LIns* ins);
			void        asm_immf4(LIns* ins);
			void        asm_ui2f(LIns* ins);
			void        asm_d2f(LIns* ins);
			void        asm_f2d(LIns* ins);
			void        asm_f2f4(LIns* ins);
			void        asm_f4comp(LIns* ins);

			void        asm_cvt2itf(LIns* ins);
			void        asm_cvt2irf(LIns* ins);
			void        asm_cvt2uitf(LIns* ins);
			void        asm_cvt2uirf(LIns* ins);
			void        asm_cvt2itd(LIns* ins);
			void        asm_cvt2ird(LIns* ins);
			void        asm_cvt2uitd(LIns* ins);
			void        asm_cvt2uird(LIns* ins);

			void        asm_cvti2f(LIns* ins);
			void        asm_cvtui2f(LIns* ins);
			void        asm_cvti2d(LIns* ins);
			void        asm_cvtui2d(LIns* ins);
			void        asm_cvtf2d(LIns* ins);

			void		asm_movi2f(LIns* ins);
			void		asm_movi2d(LIns* ins);
			void		asm_movf2d(LIns* ins);
			void		asm_movf2i(LIns* ins);
			void		asm_movd2q(LIns* ins);
			void		asm_movq2d(LIns* ins);

			void        asm_gp_copy(Register d, Register s);
			void        asm_nongp_copy(Register d, Register s);
			void        asm_call(LIns* ins, const NIns* address);
			NIns*       asm_branch(const NIns* targ);
			void        asm_branch_cond(bool branchOnFalse, LIns* cond, const NIns* targ, bool patchable, Branches* branches);
			void        asm_insert_random_nop();
			void        asm_label();
			void        asm_store_to_stack(LIns* ins, int offset, RegisterMask allow);

			void        asm_movi(LIns* ins);
			void        asm_movq(LIns* ins);
			void        asm_movf(LIns* ins);
			void        asm_movd(LIns* ins);
			void        asm_movf4(LIns* ins);

			void        reconcileExitRegs(ParameterValue* paramData, unsigned int paramCount);
			void        reconcileParamRegs(ParameterValue* paramData, unsigned int paramCount);

			// platform specific implementation (see NativeXXX.cpp file)
			void        nCreate();
			void        nDestroy();
			void        nBeginAssembly(Allocator* compilationAllocator);
			void        nEndAssembly(Allocator* compilationAllocator);

			static void nPatchBranch(NIns* branch, const NIns* target);
			static void nPatchRelativeOffset(const NIns *originalAddress, NIns *opcodeAddress);

			void        spawnValue(LIns* ins, RegisterMask allow);

			void        asm_phiReconcile(LIns* ins);
			void        asm_phiReconcileForwards(LIns* ins);
			void        asm_phiReconcileBackwards(LIns* ins);
			void        asm_phiCapture(LIns* ins);
			void		setReconcileDefaultRegAlloc(LIns* reconcileIns, PhiState* phiStateArray, size_t phiStateCount);
			void		asm_reconcileStateBackward(LIns* value, bool inReg, RegisterMask regmask, bool inAr, unsigned int arIndex);

			void        asm_spillFence(LIns* ins);
			void        asm_spillFenceAll(LIns* ins);
			void        asm_registerFence(LIns* ins);
			void        asm_registerFenceAll(LIns* ins);
			void        asm_killFence(LIns* ins);

			Register    allocReg(LIns* ins, RegisterMask allow);
			Register    allocTempReg(RegisterMask allow);
			Register    moveToReg(LIns* ins, RegisterMask allow);
			static void setResultLive(LIns* ins);
			void        asm_copy_slot(LIns* ins, LIns* insToAssign);
			void        asm_force_gen_flags(LIns* ins);
			// platform specific methods
		public:
			DECLARE_PLATFORM_ASSEMBLER()
	};

	inline int32_t arDisp(uint32_t arIndex) {
		// even on 64bit cpu's, we allocate stack area in 1byte chunks
		return int32_t(arIndex);
	}

	inline const Config& Assembler::getConfig() const {
		return _config;
	}

	inline AssemblerError Assembler::getError() const {
		return _err;
	}

	inline bool Assembler::hasError() const {
		return _err != None;
	}

#ifdef NANOJIT_VERBOSE
	inline const CompilationStats* Assembler::getCompilationStats() const {
		return &compilationStats;
	}
#endif

	inline CodeBlockList Assembler::getCodeBlockList() const {
		return _codeBlockList;
	}

	inline const EntryPointList& Assembler::getEntryPointList() const {
		return _entryPointList; 
	}
	
	inline const RelativeOffsetPatchList& Assembler::getRelativeOffsetPatchList() const {
		return _relativeOffsetPatchList;
	}

	inline const RelocationList& Assembler::getBranchRelocationList() const {
		return _branchRelocationList;
	}

	inline void Assembler::setResultLive(LIns* ins) {
		ins->setResultLive();
	}

	inline Register Assembler::findSpecificRegFor(LIns* i, Register w) {
		return findRegFor(i, rmask(w));
	}

	// Like findRegFor(), but called when the LIns is used as a pointer.  It
	// doesn't have to be called, findRegFor() can still be used.
	inline Register Assembler::getBaseReg(LIns* base, int &d, RegisterMask  allow) {

		(void) d;
		Register r =  findRegFor(base, allow);
		return r;
	}

	// Like findRegFor2(), but used for stores where the base value has the
	// same type as the stored value, eg. in asm_store32() on 32-bit platforms
	// and asm_store64() on 64-bit platforms.
	inline void Assembler::getBaseReg2(RegisterMask allowValue, LIns* value, Register&  rv,
		RegisterMask allowBase, LIns* base, Register& rb, int &d) {

		(void) d;
		findRegFor2(allowValue, value, rv, allowBase, base, rb);
	}

}
#endif // __nanojit_Assembler__
