/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nanojit.h"

//#define DOPROF
#include "vprof/vprof.h"

#ifdef FEATURE_NANOJIT


namespace nanojit
{

#if defined(NANOJIT_VERBOSE)
	static const bool verbose = false;
#else
	static const bool verbose = false;
#endif

	// Sanity checks that should remain enabled in release builds.
	#define ABORT_UNLESS(cond) do { NanoAssert(cond); if (!(cond)) VMPI_abort(); } while(0)

	CodeAlloc::CodeAlloc(CodeAllocatorImpl* impl)
		: _impl(impl)
		, heapBlockList(NULL)
		, availableBlockList(NULL)
		, totalAllocated(0)
		, bytesPerAlloc(0)
	{
		bytesPerAlloc = _impl->getChunkSizeInBytes();
	}

	CodeAlloc::~CodeAlloc() {
		reset();
	}

	void CodeAlloc::reset() {

#if defined(NANOJIT_DISABLE_CODE_FREE)
		HeapBlock* heapBlock = heapBlockList;

		while (heapBlock != NULL) {
			markHeapBlockExec(heapBlock);
			HeapBlock* nextHeapBlock = heapBlock->next();
			heapBlock = nextHeapBlock;
		}

#else
		// give all memory back to gcheap.  Assumption is that all
		// code is done being used by now.
		HeapBlock* heapBlock = heapBlockList;

		while (heapBlock != NULL) {
			_nvprof("free page", 1);
			// save next pointer (memory will become unavailable)
			HeapBlock* nextHeapBlock = heapBlock->next();

			markHeapBlockWrite(heapBlock);
			NanoAssert(heapBlock->size() == bytesPerAlloc);
			_impl->freeCodeChunk(heapBlock->buffer());
			totalAllocated -= bytesPerAlloc;
			heapBlock = nextHeapBlock;
		}

		NanoAssert(totalAllocated == 0);

		heapBlockList = NULL;
		availableBlockList.clear();
#endif
	}

	bool CodeAlloc::hasActiveAllocations() const {

#if defined(NANOJIT_DISABLE_CODE_FREE)
		return false;
#else
		const CodeBlock* availableBlock = availableBlockList.head();

		size_t totalAvailableSize = 0;
		size_t totalAllocatedSize = 0;

		while (availableBlock != NULL) {
			totalAvailableSize += availableBlock->allocatedSize();
			availableBlock = availableBlock->next();
		}

		const HeapBlock* heapBlock = heapBlockList;
		while (heapBlock != NULL) {
			totalAllocatedSize += (heapBlock->size() - sizeof(HeapBlock));
			heapBlock = heapBlock->next();
		}

		return (totalAllocatedSize != totalAvailableSize);
#endif
	}

	static int round(size_t x) {
		return (int)((x + 512) >> 10);
	}

	void CodeAlloc::getStats(size_t& total, size_t& frag_size, size_t& free_size) {
		total = 0;
		frag_size = 0;
		free_size = 0;
		int free_count = 0;

		for (const HeapBlock* heapBlock = heapBlockList; heapBlock != NULL; heapBlock = heapBlock->next()) {
			total += heapBlock->size();
		}

		for (CodeBlock* block = availableBlockList.head(); block != NULL; block = block->next()) {

			NanoAssert(block->isFree());
			NanoAssert(block->internalBufferSize() >= minAllocSize);

			free_count++;
			free_size += block->allocatedSize();
		}
	}

	void CodeAlloc::logStats() {
		size_t total, frag_size, free_size;
		getStats(total, frag_size, free_size);
		avmplus::AvmLog("code-heap: %dk free %dk fragmented %d\n",
			round(total), round(free_size), frag_size);
	}

	CodeBlock* CodeAlloc::alloc() {

		if (availableBlockList.isEmpty()) {
			// no free mem, get more
			addMem();
		}

		// grab a block
		markHeapBlockWrite(availableBlockList.head()->terminator());

		CodeBlock* block = availableBlockList.removeFirstBlockFromList();
		block->markUsed();

		NanoAssert(block->internalBufferSize() >= minAllocSize);

		if (verbose)
			avmplus::AvmLog("CodeAlloc(%p).alloc %p-%p %d\n", this, block->internalBufferStart(), block->internalBufferEnd(), int(block->internalBufferSize()));
		
		debug_only(sanityCheck();)

		return block;
	}

	CodeBlock* CodeAlloc::allocFixedSize(size_t a_iSize) {

		// size cannot be greater than allocation chunk size
		NanoAssert(a_iSize <= (bytesPerAlloc - CodeBlock::sizeOfBlockHeader() - sizeof(HeapBlock)));

		CodeBlock* codeBlock = availableBlockList.head();
		CodeBlock* parentBlock = NULL;

		while (codeBlock != NULL) {

			if (codeBlock->codeSize() >= a_iSize) {
				markHeapBlockWrite(codeBlock->terminator());
				parentBlock = availableBlockList.findParent(codeBlock);
				if (parentBlock == NULL) {
					availableBlockList.removeFirstBlockFromList();
				} else {
					markHeapBlockWrite(parentBlock->terminator());
					availableBlockList.removeNextBlockFromList(parentBlock);
				}
				break;
			}

			codeBlock = codeBlock->next();
		}

		if (codeBlock == NULL) {
			// allocate a new chunk big enough for the allocation
			// the chunk will be inserted at the beginning of the list
			addMem();
			codeBlock = availableBlockList.removeFirstBlockFromList();
			NanoAssert(codeBlock->codeSize() >= a_iSize);
		}

		codeBlock->markUsed();

		NIns* blockEnd = reinterpret_cast<NIns*>(reinterpret_cast<uint8_t*>(codeBlock->internalBufferEnd()) - a_iSize);
		codeBlock = addRemainder(codeBlock, blockEnd);

		if ((parentBlock != NULL) && (parentBlock->terminator() != codeBlock->terminator())) {
			markHeapBlockExec(parentBlock->terminator());
		}

		debug_only(sanityCheck();)
		NanoAssert(codeBlock->internalBufferSize() >= a_iSize);

		return codeBlock;
	}

	void CodeAlloc::free(CodeBlock* block) {

#if defined(NANOJIT_DISABLE_CODE_FREE)
		markHeapBlockExec(block->terminator());
#else
		NanoAssert(heapBlockList != NULL);
		CodeBlock* currentBlock = block;
		CodeBlock* parentBlock = NULL;

		if (verbose)
			avmplus::AvmLog("free %p-%p %d\n", currentBlock->internalBufferStart(), currentBlock->internalBufferEnd(), (int)currentBlock->internalBufferSize());

		NanoAssert(currentBlock->internalBufferSize() >= minAllocSize);
		NanoAssert(!currentBlock->isFree());

		markHeapBlockWrite(currentBlock->terminator());

		bool bAlreadyInAvaiblableList = false;

		currentBlock->markFree();
		
#if defined(NANOJIT_DEBUG)
		::memset(currentBlock->internalBufferStart(), 0xCE, currentBlock->internalBufferSize());
#endif

		// coalesce with lower
		{
			CodeBlock* lower = currentBlock->lower();
			if ((lower != NULL) && lower->isFree()) {

				NanoAssert(lower->terminator() == currentBlock->terminator());

				CodeBlock::mergeCurrentWithUpper(lower);
				currentBlock = lower;
				NanoAssert(currentBlock->internalBufferSize() >= minAllocSize);
				bAlreadyInAvaiblableList = true;
			}
		}

		// coalesce with upper
		{
			CodeBlock* upper = currentBlock->upper();
			if ((upper != NULL) && upper->isFree()) {

				NanoAssert(upper->terminator() == currentBlock->terminator());
				NanoAssert(availableBlockList.isMember(upper));

				parentBlock = availableBlockList.findParent(upper);
				if (parentBlock == NULL) {
					availableBlockList.removeFirstBlockFromList();
				} else {
					markHeapBlockWrite(parentBlock->terminator());
					availableBlockList.removeNextBlockFromList(parentBlock);
				}

				upper->markFree();

				CodeBlock::mergeCurrentWithUpper(currentBlock);
				NanoAssert(currentBlock->internalBufferSize() >= minAllocSize);
			}
		}

		//
		NanoAssert(!currentBlock->lower() || !currentBlock->lower()->isFree());
		NanoAssert(!currentBlock->upper() || !currentBlock->upper()->isFree());
		NanoAssert(currentBlock->internalBufferSize() >= minAllocSize);

		if (!bAlreadyInAvaiblableList)
			availableBlockList.addBlockToList(currentBlock);



		markHeapBlockExec(currentBlock->terminator());
		if (parentBlock != NULL)
			markHeapBlockExec(parentBlock->terminator());

		NanoAssert(!availableBlockList.isEmpty());
#endif // NANOJIT_DISABLE_CODE_FREE
	}

	void CodeAlloc::freeList(CodeBlockList* codeBlockList) {

#if defined(NANOJIT_DEBUG)
		NanoAssert(testIsAllExec());
#endif 

		while (!codeBlockList->isEmpty()) {
			markHeapBlockWrite(codeBlockList->head()->terminator());
			CodeBlock* b = codeBlockList->removeFirstBlockFromList();
			free(b);
		}

#if defined(NANOJIT_DEBUG)
		sanityCheck();
		NanoAssert(testIsAllExec());
#endif
	}

	void CodeAlloc::flushBlockIcache(CodeBlock* codeBlock) {
		_impl->flushICache(codeBlock->internalBufferStart(), codeBlock->internalBufferSize());
	}

	void CodeAlloc::flushListIcache(CodeBlockList* codeBlockList) {
		for (CodeBlock* b = codeBlockList->head(); b != NULL; b = b->next()) {
			flushBlockIcache(b);
		}
	}

	void CodeAlloc::addMem() {
		// allocations never fail, and memory is supposed to be writable
		void *mem = _impl->allocCodeChunk(); 
		totalAllocated += bytesPerAlloc;
		NanoAssert(mem != NULL); // see allocCodeChunk contract in CodeAlloc.h
		_nvprof("alloc page", uintptr_t(mem)>>12);

#if defined(NANOJIT_DEBUG)
		// temp debug: INT 3 instruction
		::memset(mem, 0xCC, bytesPerAlloc);
#endif

		HeapBlock* heapBlock = HeapBlock::setupHeapBlockFromBuffer(mem, bytesPerAlloc, heapBlockList);
		heapBlockList = heapBlock;

		CodeBlock* newBlock = heapBlock->lowestCodeBlock();
		availableBlockList.addBlockToList(newBlock);

		debug_only(sanityCheck();)
	}

	/**
	 * split a block by freeing the hole in the middle defined by [holeStart,holeEnd),
	 * and adding the used prefix and suffix parts to the blocks CodeList.
	 */
	CodeBlock* CodeAlloc::addRemainder(CodeBlock* block, NIns* codeStart) {

		NanoAssert(!block->terminator()->isExec());
		NIns const* end = block->internalBufferEnd();

		CodeBlock* resultBlock = NULL;

		NanoAssert(block->internalBufferStart() < end);
		NanoAssert((block->internalBufferStart() <= codeStart) && (end >= codeStart));

		if (codeStart == end) {
			NanoAssert((block->lower() == NULL) || (!block->lower()->isFree()));
			NanoAssert((block->upper() == NULL) || (!block->upper()->isFree()));

			// empty block. free whole start-end range
			availableBlockList.addBlockToList(block);
			resultBlock = NULL;

		} else if (CodeBlock::canSplitBlock(minAllocSize, block, codeStart)) {

			CodeBlock* splittedBlock = CodeBlock::splitBlock(minAllocSize, block, codeStart);

			splittedBlock->markUsed();
			block->markFree();

			NanoAssert(splittedBlock->lower() == block);
			NanoAssert((splittedBlock->upper() == NULL) || (!splittedBlock->upper()->isFree()));

			NanoAssert(block->upper() == splittedBlock);
			NanoAssert((block->lower() == NULL) || (!block->lower()->isFree()));

			NanoAssert(splittedBlock->internalBufferSize() >= minAllocSize);
			NanoAssert(block->internalBufferSize() >= minAllocSize);

			availableBlockList.addBlockToList(block);
			resultBlock = splittedBlock;

		} else {
			// the remaining hole is too small to make a new free block and a new used block. just keep
			// the whole original block and don't free anything.
			block->markUsed();
			resultBlock = block;
		}

		debug_only(sanityCheck();)

		return resultBlock;
	}

#ifdef PERFM
	// This method is used only for profiling purposes.
	// See CodegenLIR::emitMD() in Tamarin for an example.

	size_t CodeAlloc::size(const CodeList* blocks) {
		size_t size = 0;
		for (const CodeList* b = blocks; b != 0; b = b->_next)
			size += int((uintptr_t)b->end - (uintptr_t)b);
		return size;
	}
#endif

	size_t CodeAlloc::size() {
		return totalAllocated;
	}

	// check that all block neighbors are correct
	#if defined(NANOJIT_DEBUG)
	bool CodeAlloc::testIsAllExec() {
		for (HeapBlock* hb = heapBlockList; hb != 0; hb = hb->next()) {
			if (!hb->isExec())
				return false;
		}
		return true;
	}

	void CodeAlloc::sanityCheck() {

		size_t totalHeapBlockBytes = 0;
		size_t totalCodeBlockBytes = 0;

		for (HeapBlock* hb = heapBlockList; hb != 0; hb = hb->next()) {

			totalHeapBlockBytes += hb->size();
			NanoAssert(totalHeapBlockBytes <= totalAllocated);
			NanoAssert(hb->lowestCodeBlock() != NULL);

			totalCodeBlockBytes += sizeof(HeapBlock);

			for (const CodeBlock* b = hb->lowestCodeBlock(); b != 0; b = b->upper()) {

				NanoAssert((b->upper() == NULL) || (b->upper()->lower() == b));
				NanoAssert((b->lower() == NULL) || (b->lower()->upper() == b));

				totalCodeBlockBytes += b->allocatedSize();

				NanoAssert(b->internalBufferSize() >= minAllocSize);
				NanoAssert(totalCodeBlockBytes <= totalAllocated);
			}
			if (true) {
				NanoAssert(hb->size() == bytesPerAlloc);
				bool b = _impl->checkChunkMark(hb->buffer(), hb->isExec());
				NanoAssertMsg(b, "Chunk access mode differs from that expected");
			}
		}

		NanoAssert(totalHeapBlockBytes == totalAllocated);
		NanoAssert(totalCodeBlockBytes == totalAllocated);

		for (CodeBlock* avail = this->availableBlockList.head(); avail; avail = avail->next()) {
			NanoAssert(avail->isFree() && avail->internalBufferSize() >= minAllocSize);
		}
		

		#if 0 //CROSS_CHECK_FREE_LIST
		for(HeapBlock* term = heapBlockList; term; term = term->next()) {
			for(const CodeBlock* cb = term->lowestCodeBlock(); cb != NULL; cb = cb->lower()) {
				NanoAssert(cb->internalBufferSize() >= minAllocSize);

				if (cb->isFree()) {
					bool foundInAvailableList = false;
					for (const CodeBlock* avail = availableBlockList.head(); !foundInAvailableList && avail; avail = avail->next()) {
						foundInAvailableList = (avail == cb);
					}
					NanoAssert(foundInAvailableList);
				}
			}
		}

		for (const CodeBlock* avail = availableBlockList.head(); avail; avail = avail->next()) {
			bool foundInHeapblocks = false;
			for(HeapBlock* term = heapBlockList; !foundInHeapblocks && term; term = term->next()) {
				for(const CodeBlock* cb = term->lowestCodeBlock(); !foundInHeapblocks && cb; cb = cb->upper()) {
					foundInHeapblocks = (cb == avail);
				}
			}
			NanoAssert(foundInHeapblocks);
		}
		#endif /* CROSS_CHECK_FREE_LIST */
	}
	#endif

	// Loop through a list of blocks marking the chunks executable.  If we encounter
	// multiple blocks in the same chunk, only the first block will cause the
	// chunk to become executable, the other calls will no-op (isExec flag checked)
	void CodeAlloc::markListExec(CodeBlockList* blockList) {
		for (CodeBlock* b = blockList->head(); b != 0; b = b->next()) {
			markHeapBlockExec(b->terminator());
		}
	}

	// Variant of markExec(CodeList*) that walks all heapblocks (i.e. chunks) marking
	// each one executable.   On systems where bytesPerAlloc is low (i.e. have lots
	// of elements in the list) this can be expensive.
	void CodeAlloc::markAllExec() {
		for (HeapBlock* hb = heapBlockList; hb != NULL; hb = hb->next()) {
			markHeapBlockExec(hb);
		}
	}

	// make an entire chunk executable
	void CodeAlloc::markHeapBlockExec(HeapBlock* term) {
		if (!term->isExec()) {
			term->markExec();
			NanoAssert(term->size() == bytesPerAlloc);
			_impl->markCodeChunkExec(term->buffer());
		}
	}


	// Loop through a list of blocks marking the chunks writable. If we encounter
	// multiple blocks in the same chunk, only the first block will cause the
	// chunk to become executable, the other calls will no-op (isExec flag checked)
	void CodeAlloc::markListWrite(CodeBlockList* blockList) {
		for (CodeBlock* b = blockList->head(); b != 0; b = b->next()) {
			markHeapBlockWrite(b->terminator());
		}
	}

	// Variant of markExec(CodeList*) that walks all heapblocks (i.e. chunks) marking
	// each one executable.   On systems where bytesPerAlloc is low (i.e. have lots
	// of elements in the list) this can be expensive.
	void CodeAlloc::markAllWrite() {
		for (HeapBlock* hb = heapBlockList; hb != NULL; hb = hb->next()) {
			markHeapBlockWrite(hb);
		}
	}

	void CodeAlloc::markHeapBlockWrite(HeapBlock* term) {
		if (term->isExec()) {
			NanoAssert(term->size() == bytesPerAlloc);
			_impl->markCodeChunkWrite(term->buffer());
			term->markWritable();
		}
	}
}
#endif // FEATURE_NANOJIT
