/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nanojit.h"

NanoStaticAssert(nanojit::LIR_sentinel <= 0x100);
//NanoStaticAssert(sizeof(nanojit::LInsCompilationData) == 8);
//NanoStaticAssert(sizeof(nanojit::LInsData) == 24);
//NanoStaticAssert(sizeof(nanojit::LIns) == (8 + 24 + 8));
//
namespace nanojit
{
	using namespace avmplus;
	#ifdef FEATURE_NANOJIT
		
	const uint8_t repKinds[] = {
#define OP___(op, repKind, retType, isCse) \
		LRK_##repKind,
#include "LIRopcode.tbl"
#undef OP___
		0
	};

	const LTy retTypes[] = {
#define OP___(op, repKind, retType, isCse) \
		LTy_##retType,
#include "LIRopcode.tbl"
#undef OP___
		LTy_V
	};

	const uint8_t insSizes[] = {
#define OP___(op, repKind, retType, isCse) \
		sizeof(LIns##repKind),
#include "LIRopcode.tbl"
#undef OP___
		0
	};

	const int8_t isCses[] = {
#define OP___(op, repKind, retType, isCse) \
		isCse,
#include "LIRopcode.tbl"
#undef OP___
		0
	};

	LIns::LIns() 
	: compilationData()
	, arGroup(0)
	, op(LIR_sentinel)
	, _allowFolding(true)
	, _genFlags(0) {

	}


	LIns::~LIns() {

	}

	void LIns::initLInsPhi(LOpcode opcode, PhiData* phiData, LIns** values, size_t count, LIns* reconcileNodeNext) {
		initSharedFields(opcode);
		getLInsPhi()->phiData			= phiData;
		getLInsPhi()->reconcileNodeNext = reconcileNodeNext;
		getLInsPhi()->phiData = NULL;
		getLInsPhi()->values = NULL;
		getLInsPhi()->count = 0;
		phiSetValues(values, count);
		phiSetData(phiData);
		NanoAssert(isLInsPhi());
	}

	LOpcode arithOpcodeD2I(LOpcode op) {
		switch (op) {
		case LIR_negd:      return LIR_negi;
		case LIR_addd:      return LIR_addi;
		case LIR_subd:      return LIR_subi;
		case LIR_muld:      return LIR_muli;
		default:            NanoAssert(0); return LIR_sentinel;
		}
	}

#ifdef NANOJIT_64BIT
	LOpcode cmpOpcodeI2Q(LOpcode op) {
		switch (op) {
		case LIR_eqi:       return LIR_eqq;
		case LIR_nei:       return LIR_neq;
		case LIR_lti:       return LIR_ltq;
		case LIR_gti:       return LIR_gtq;
		case LIR_lei:       return LIR_leq;
		case LIR_gei:       return LIR_geq;
		case LIR_ltui:      return LIR_ltuq;

		case LIR_gtui:      return LIR_gtuq;
		case LIR_leui:      return LIR_leuq;
		case LIR_geui:      return LIR_geuq;
		default:            NanoAssert(0); return LIR_sentinel;
		}
	}
#endif

	LOpcode cmpOpcodeD2I(LOpcode op) {
		switch (op) {
		case LIR_eqd:   return LIR_eqi;
		case LIR_ned:   return LIR_nei;
		case LIR_ltd:   return LIR_lti;
		case LIR_gtd:   return LIR_gti;
		case LIR_led:   return LIR_lei;
		case LIR_ged:   return LIR_gei;
		default:        NanoAssert(0); return LIR_skip;
		}
	}

	LOpcode cmpOpcodeD2UI(LOpcode op) {
		switch (op) {
		case LIR_eqd:    return LIR_eqi;
		case LIR_ned:    return LIR_nei;
		case LIR_ltd:    return LIR_ltui;
		case LIR_gtd:    return LIR_gtui;
		case LIR_led:    return LIR_leui;
		case LIR_ged:    return LIR_geui;
		default:        NanoAssert(0); return LIR_skip;
		}
	}

	
	void LIns::overwriteWithSkip(LIns* skipTo) {
		// Ensure the instruction is at least as big as a LIR_skip.
		NanoAssert(insSizes[opcode()] >= insSizes[LIR_skip]);
		initLInsSk(skipTo);
	}


	// This is an ugly workaround for an apparent compiler
	// bug; in VC2008, compiling with optimization on
	// will produce spurious errors if this code is inlined
	// into ExprFilter::ins2(). See https://bugzilla.mozilla.org/show_bug.cgi?id=538504
	inline double do_join(int32_t c1, int32_t c2) {
		union {
			double d;
			uint64_t u64;
		} u;
		u.u64 = uint32_t(c1) | uint64_t(c2)<<32;
		return u.d;
	}




	using namespace avmplus;


#ifdef NANOJIT_VERBOSE

	void LirNameMap::addNameWithSuffix(LIns* ins, const char *name, int suffix,
									   bool ignoreOneSuffix) {
		NanoAssert(name != NULL);
		NanoAssert(!names.containsKey(ins));
		const int N = 100;
		char name2[N];
		if (suffix == 1 && ignoreOneSuffix) {
			VMPI_snprintf(name2, N, "%s", name);                // don't add '1' suffix
		} else if ((name[0] != '\0') && VMPI_isdigit(name[VMPI_strlen(name)-1])) {
			VMPI_snprintf(name2, N, "%s_%d", name, suffix);     // use '_' to avoid confusion
		} else {
			VMPI_snprintf(name2, N, "%s%d", name, suffix);      // normal case
		}

		char *copy = new (alloc) char[VMPI_strlen(name2)+1];
		VMPI_strcpy(copy, name2);
		Entry *e = new (alloc) Entry(copy);
		names.put(ins, e);
	}

	void LirNameMap::addName(LIns* ins, const char* name) {
		// The lookup may succeed, ie. we may already have a name for this
		// instruction.  This can happen because of CSE.  Eg. if we have this:
		//
		//   ins = addName("foo", insImmI(0))
		//
		// that assigns the name "foo1" to 'ins'.  If we later do this:
		//
		//   ins2 = addName("foo", insImmI(0))
		//
		// then CSE will cause 'ins' and 'ins2' to be equal.  So 'ins2'
		// already has a name ("foo1") and there's no need to generate a new
		// name "foo2".
		//
		if (!names.containsKey(ins)) {
			Str* str = new (alloc) Str(alloc, name);
			int suffix = namecounts.add(*str);
			addNameWithSuffix(ins, name, suffix, /*ignoreOneSuffix*/true);
		}
	}

	const char* LirNameMap::createName(LIns* ins) {
		NanoAssert(ins != NULL);
		if (!names.containsKey(ins))
		{
			addNameWithSuffix(ins, getLirName(ins->opcode()),   lircounts.add(ins->opcode()),
								/*ignoreOneSuffix*/false);
		}

		return names.get(ins)->name;
	}

	const char* LirNameMap::lookupName(LIns* ins) {
		Entry* e = names.get(ins);
		return e ? e->name : NULL;
	}


#endif



#endif /* FEATURE_NANOJIT */

#if defined(NANOJIT_VERBOSE)
	AddrNameMap::AddrNameMap(Allocator* a)
		: allocator(a), names(a) {}

	void AddrNameMap::addAddrRange(const void *p, size_t size, size_t align, const char *name) {
		if (!this || names.containsKey(p))
			return;
		char* copy = new (allocator) char[VMPI_strlen(name)+1];
		VMPI_strcpy(copy, name);
		Entry *e = new (allocator) Entry(copy, size << align, align);
		names.put(p, e);
	}

	void AddrNameMap::lookupAddr(void *p, char*& name, int32_t& offset) {
		const void *start = names.findNear(p);
		if (start) {
			Entry *e = names.get(start);
			const void *end = (const char*)start + e->size;
			if (p == start) {
				name = e->name;
				offset = 0;
			}
			else if (p > start && p < end) {
				name = e->name;
				offset = int32_t(intptr_t(p)-intptr_t(start)) >> e->align;
			}
			else {
				name = NULL;
				offset = 0;
			}
		} else {
			name = NULL;
			offset = 0;
		}
	}

#endif // NANOJIT_VERBOSE

#ifdef FEATURE_NANOJIT
#ifdef NANOJIT_DEBUG


	ValidateReader::ValidateReader(LirFilter* in) : LirFilter(in) {}

	LIns* ValidateReader::read() {
		LIns *ins = in->read();
		switch (ins->opcode()) {
		case LIR_jt:
		case LIR_jf:
		case LIR_j:
			NanoAssert((ins->branchTarget() != NULL) && (ins->branchTarget()->isLInsLabel()));
			break;

		default:
			;
		}
		return ins;
	}

#endif
#endif

}
