
#include "nanojit.h"

namespace nanojit
{

LIns* getPreviousElement(LIns* a_pLIns)
{
	LInsElement* pElement = LInsElement::getElement(a_pLIns);
	IntrusiveLinkedList::ReverseIterator it = IntrusiveLinkedList::getReverseIteratorAt(pElement);
	it.moveNext();
	LInsElement* pPreviousElement = it.getElement<LInsElement>();

	return pPreviousElement->data();
}


LirChainWriter::LirChainWriter()
: _storage()
, _allocator(NULL)
#if defined(NANOJIT_DEBUG)
, m_iWasPeeked(0)
#endif
{
}

LirChainWriter::~LirChainWriter()
{
}

void LirChainWriter::initialize(Allocator* allocator)
{
	_allocator = allocator;

	_storage.initialize(allocator);
}

void LirChainWriter::shutdown()
{
	_storage.shutdown();

	_allocator = NULL;
}

void LirChainWriter::start()
{
	clear();

	ins0(nanojit::LIR_start);
}

void LirChainWriter::clear()
{
	_storage.clear();
}

void LirChainWriter::end()
{
#if defined(NANOJIT_DEBUG)
	NanoAssert(m_iWasPeeked == 0);
#endif
}

LIns* LirChainWriter::ins0(LOpcode v)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp0(v);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::ins1(LOpcode v, LIns* a)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp1(v, a);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::ins2(LOpcode v, LIns* a, LIns* b)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp2(v, a, b);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::ins3(LOpcode v, LIns* a, LIns* b, LIns* c)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp3(v, a, b, c);

	return pIns;
}

LIns* LirChainWriter::insLabel()
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsLabel();

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insPhi(LOpcode op, PhiData* phiData, LIns** values, size_t count, LIns* reconcileNodeNext)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsPhi(op, phiData, values, count, reconcileNodeNext);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insSpillFence(LIns** valueArray, size_t count)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsValueFence(valueArray, count);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insSpillFenceAll()
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp0(LIR_spillfenceall);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


LIns* LirChainWriter::insRegisterFence(LIns** valueArray, size_t count)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsRegisterFence(valueArray, count);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insRegisterFenceAll()
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp0(LIR_regfenceall);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insKillFence(LIns** valueArray, size_t count)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsKillFence(valueArray, count);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


LIns* LirChainWriter::insBranch(LOpcode v, LIns* condition, LIns* to)
{
	NanoAssert((v == LIR_j && !condition) || ((v == LIR_jf || v == LIR_jt) && condition));
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsBranch(v, condition);

	if (to != NULL)
		pIns->branchSetTarget(to);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insEntry(RegisterMask saveMask)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsEntry(saveMask);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


LIns* LirChainWriter::insExit(ParameterValue* values, unsigned int count, RegisterMask restoreMask)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsExit(values, count, restoreMask);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


LIns* LirChainWriter::insParam(LOpcode op, Register reg)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsP(op, reg);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insImmI(int32_t imm)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsI(imm);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


#ifdef NANOJIT_64BIT
LIns* LirChainWriter::insImmQ(int64_t imm)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsQ(imm);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}
#endif

LIns* LirChainWriter::insImmF(float f)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsF(f);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insImmF4(const float* f4)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsF4(f4);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insImmD(double d)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsD(d);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insLoad(LOpcode op, LIns* base, int32_t d, AccSet accSet, LoadQual loadQual)
{
	if (isS16(d))
	{
		LInsElement*    pElement    = _storage.insertNewElement();
		LIns*           pIns        = pElement->data();
		
		pIns->reset();
		pIns->initLInsLd(op, base, d, accSet, loadQual);

#if defined(NANOJIT_DEBUG)
		if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

		return pIns;
	} 
	else
	{
		// If the displacement is more than 16 bits, put it in a separate instruction.
		// Note that CseFilter::insLoad() also does this, so this will
		// only occur if CseFilter has been removed from the pipeline.
		return insLoad(op, ins2(LIR_addp, base, insImmP(reinterpret_cast<void*>(d))), 0, accSet, loadQual);
	}
}

LIns* LirChainWriter::insStore(LOpcode op, LIns* value, LIns* base, int32_t d, AccSet accSet)
{
	if (isS16(d))
	{
		LInsElement*    pElement    = _storage.insertNewElement();
		LIns*           pIns        = pElement->data();

		pIns->reset();
		pIns->initLInsSt(op, value, base, d, accSet);

#if defined(NANOJIT_DEBUG)
		if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

		return pIns;
	} 
	else
	{
		// If the displacement is more than 16 bits, put it in a separate instruction.
		return insStore(op, value, ins2(LIR_addp, base, insImmP(reinterpret_cast<void*>(d))), 0, accSet);
	}
}

LIns* LirChainWriter::insCall(LOpcode op, CallParameters* parameters, const void* address)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsCall(op, parameters, address);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insComment(const char* str)
{
	std::size_t iLength = std::strlen(str) + 1;
	char* sMessage = (char*)_allocator->alloc(iLength);
	std::memcpy(sMessage, str, iLength);

	return ins1(LIR_comment, (LIns*)sMessage);
}

LIns* LirChainWriter::insSkip(LIns* skipTo)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsSk(skipTo);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insSwz(LIns* a, uint8_t mask)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->initLInsOp1b(LIR_swzf4, a, mask);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}

LIns* LirChainWriter::insCopy(LIns* insToCopy)
{
	LInsElement*    pElement    = _storage.insertNewElement();
	LIns*           pIns        = pElement->data();

	pIns->reset();
	pIns->duplicateFrom(insToCopy);

#if defined(NANOJIT_DEBUG)
	if (m_iWasPeeked > 0) m_iWasPeeked--;
#endif

	return pIns;
}


void LirChainWriter::moveBefore(LIns* insToMove, LIns* insTarget)
{
	LInsElement* elementToMove = LInsElement::getElement(insToMove);
	LInsElement* elementTarget = LInsElement::getElement(insTarget);

#if defined(NANOJIT_DEBUG)
	NanoAssert(m_iWasPeeked == 0);
#endif

	_storage.moveBefore(elementToMove, elementTarget);
}


void LirChainWriter::moveAfter(LIns* insToMove, LIns* insTarget)
{
	LInsElement* elementToMove = LInsElement::getElement(insToMove);
	LInsElement* elementTarget = LInsElement::getElement(insTarget);

#if defined(NANOJIT_DEBUG)
	NanoAssert(m_iWasPeeked == 0);
#endif

	_storage.moveAfter(elementToMove, elementTarget);
}

}
