/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef __nanojit_h__
#define __nanojit_h__

#include "AvmPlus.h"
#include "NanojitPortability.h"
#include "NanojitIntrinsics.h"
#include "njcpudetect.h"

#ifdef FEATURE_NANOJIT

#if defined(_DEBUG) || !(defined(NDEBUG))
#define NANOJIT_DEBUG
#endif

#if !defined(NANOJIT_ENABLE_ASSERT) && defined(NANOJIT_DEBUG)
#define NANOJIT_ENABLE_ASSERT
#endif

namespace nanojit
{
	/**
	 * -------------------------------------------
	 * START AVM bridging definitions
	 * -------------------------------------------
	 */
	const uint32_t MAXARGS = 4;

	#define NanoUnusedVariable(var)         do { (void)var; } while (false)

	#if defined(NANOJIT_ENABLE_ASSERT)

		#define __NanoAssertMsgf(a, file_, line_, f, ...)\
			if (!(a)) {\
				avmplus::AvmLog("Assertion failure: " f "%s (%s:%d)\n", __VA_ARGS__, #a, file_, line_); \
				avmplus::AvmAssertFail("");\
			}

		#define _NanoAssertMsgf(a, file_, line_, f, ...)   __NanoAssertMsgf(a, file_, line_, f, __VA_ARGS__)

		#define NanoAssertMsgf(a,f,...)     do { __NanoAssertMsgf(a, __FILE__, __LINE__, f ": ", __VA_ARGS__); } while (0)
		#define NanoAssertMsg(a,m)          do { __NanoAssertMsgf(a, __FILE__, __LINE__, "\"%s\": ", m); } while (0)
		#define NanoAssert(a)               do { __NanoAssertMsgf(a, __FILE__, __LINE__, "%s", ""); }   while (0)
		#define NanoBreak()                 do { avmplus::AvmBreak(); } while(false)
	#else
		#define NanoAssertMsgf(a,f,...)     do { } while (0) /* no semi */
		#define NanoAssertMsg(a,m)          do { } while (0) /* no semi */
		#define NanoAssert(a)               do { } while (0) /* no semi */
		#define NanoBreak()                 do { } while (0) /* no semi */
	#endif

	/*
	 * Sun Studio C++ compiler has a bug
	 * "sizeof expression not accepted as size of array parameter"
	 * The bug number is 6688515. It is not public yet.
	 * Turn off this assert for Sun Studio until this bug is fixed.
	 */
	#ifdef __SUNPRO_CC
		#define NanoStaticAssert(condition)
	#else
		#define NanoStaticAssert(condition) \
			extern void nano_static_assert(int arg[(condition) ? 1 : -1])
	#endif


	/**
	 * -------------------------------------------
	 * END AVM bridging definitions
	 * -------------------------------------------
	 */
}


#ifdef AVMPLUS_VERBOSE
	#define NANOJIT_VERBOSE 1
#endif

#if defined(NANOJIT_VERBOSE)
	#include <stdio.h>
	#define verbose_outputf            if (_logc->getLcBits() & LC_Native) \
										Assembler::outputf
	#define verbose_only(...)        __VA_ARGS__
#else
	#define verbose_outputf
	#define verbose_only(...)
#endif /*NANOJIT_VERBOSE*/

#ifdef NANOJIT_DEBUG
	#define debug_only(x)				x
#else
	#define debug_only(x)
#endif /* NANOJIT_DEBUG */

#define isS8(i)  ( int32_t(i) == int8_t(i) )
#define isU8(i)  ( int32_t(i) == uint8_t(i) )
//### FIXME: Revisit these for types of argument and sign-correctness.
//### Should make them inlines.
#define isU12(i) ( ((i) & 0xfff) == (i) )
#define isS16(i) ( int32_t(i) == int16_t(i) )
#define isU16(i) ( int32_t(i) == uint16_t(i) )
#define isS20(i) ( (int32_t((i)<<12)>>12) == (i) )
#define isS21(i) ( (int32_t((i)<<11)>>11) == (i) )
#define isS24(i) ( (int32_t((i)<<8)>>8) == (i) )
#define isS25(i) ( (int32_t((i)<<7)>>7) == (i) )

static inline bool isS32(intptr_t i) {
	return int32_t(i) == i;
}

static inline bool isU32(uintptr_t i) {
	return uint32_t(i) == i;
}

#define alignTo(x,s)        ((((uintptr_t)(x)))&~(((uintptr_t)s)-1))
#define alignUp(x,s)        ((((uintptr_t)(x))+(((uintptr_t)s)-1))&~(((uintptr_t)s)-1))

#define NANOJIT_MIN(x, y) ((x) < (y) ? (x) : (y))
#define NANOJIT_MAX(x, y) ((x) > (y) ? (x) : (y))

namespace nanojit
{
	// Returns the index of the most significant bit that is set.
	inline unsigned msbSet32(uint32_t x) {
		NanoAssert(x != 0);
		return 31 - countLeadingZeroes32(x);
	}

	// Returns the index of the least significant bit that is set.
	inline unsigned lsbSet32(uint32_t x) {
		NanoAssert(x != 0);
		return countTrailingZeroes32(x);
	}

	// Returns the index of the most significant bit that is set.
	inline unsigned msbSet64(uint64_t x) {
		NanoAssert(x != 0);
		return 63 - countLeadingZeroes64(x);
	}

	// Returns the index of the least significant bit that is set.
	inline unsigned lsbSet64(uint64_t x) {
		NanoAssert(x != 0);
		return countTrailingZeroes64(x);
	}
} // namespace nanojit

// -------------------------------------------------------------------
// START debug-logging definitions
// -------------------------------------------------------------------

/* Debug printing stuff.  All Nanojit and jstracer debug printing
   should be routed through LogControl::printf.  Don't use
   ad-hoc calls to printf, fprintf(stderr, ...) etc.

   Similarly, don't use ad-hoc getenvs etc to decide whether or not to
   print debug output.  Instead consult the relevant control bit in
   LogControl::lcbits in the LogControl object you are supplied with.
*/

# if defined(__GNUC__)
# define PRINTF_CHECK(x, y) __attribute__((format(__printf__, x, y)))
# else
# define PRINTF_CHECK(x, y)
# endif

namespace nanojit {

	// LogControl, a class for controlling and routing debug output

	enum LC_Bits {
		/* Output control bits for Nanojit code.  Only use bits 15
		   and below, so that callers can use bits 16 and above for
		   themselves. */
		// TODO: add entries for the writer pipeline
		LC_FragProfile      = 1<<8, // collect per-frag usage counts
		LC_Liveness         = 1<<7, // show LIR liveness analysis
		LC_ReadLIR          = 1<<6, // as read from LirBuffer
		LC_AfterDCE         = 1<<4, // after dead code elimination
		LC_Bytes            = 1<<3, // byte values of native instruction
		LC_Native           = 1<<2, // final native code
		LC_RegAlloc         = 1<<1, // stuff to do with reg alloc
		LC_Activation       = 1<<0  // enable printActivationState
	};


	class LogControl
	{
		uint32_t lcbits;

	public:
		LogControl() : lcbits(0) {}
		virtual ~LogControl() { lcbits = 0; }

		virtual void printf(const char* format, ...) PRINTF_CHECK(2,3) = 0;
		virtual void flush() = 0;

		inline void setLcBits(uint32_t bits) { lcbits = bits; }
		inline uint32_t getLcBits() { return lcbits; }
	};
}

// -------------------------------------------------------------------
// END debug-logging definitions
// -------------------------------------------------------------------


#include "utils/IntrusiveLinkedList.h"
#include "njconfig.h"
#include "Allocator.h"
#include "Containers.h"
#include "Native.h"
#include "HeapBlock.h"
#include "CodeBlock.h"
#include "CodeBlockList.h"
#include "CodeAlloc.h"
#include "LIR.h"
#include "LirInstPrinter.h"
#include "LirFilter.h"
#include "LirStorage.h"
#include "LirChainWriter.h"
#include "LirOptimizationPass.h"
#include "ExprLirOptimizationPass.h"
#include "RegAlloc.h"
#include "StackAllocator.h"
#include "Assembler.h"

#endif // FEATURE_NANOJIT
#endif // __nanojit_h__
